import os
import numpy as np

from pydmclab.core.comp import CompTools
from pydmclab.core.energies import ChemPots
from pydmclab.utils.handy import read_json, write_json

from pydmclab.core.hulls import MixingHull


class VoltageCurve(object):
    def __init__(
        self,
        input_energies,
        mu_Li="dmc-r2scan",
        energy_key="E_per_at",
        intercalating_ion="Li",
        charged_composition="FeP2S6",
        discharged_composition="Li2FeP2S6",
        basis_ion=("Fe", 1),
    ):
        self.input_energies = input_energies
        if isinstance(mu_Li, str):
            mu_Li = ChemPots(
                functional=mu_Li.split("-")[1], standard=mu_Li.split("-")[0]
            ).chempots["Li"]
        self.mu_Li = mu_Li

        self.energy_key = energy_key
        self.intercalating_ion = intercalating_ion
        self.charged_composition = CompTools(charged_composition).clean
        self.discharged_composition = CompTools(discharged_composition).clean
        self.basis_ion = basis_ion

    @property
    def z(self):
        ion = self.intercalating_ion
        if ion in ["Li", "Na", "K"]:
            return 1
        elif ion in ["Mg", "Ca", "Zn"]:
            return 2
        elif ion in ["Al"]:
            return 3
        else:
            raise ValueError("ion not recognized")

    @property
    def mixing_results(self):
        hull = MixingHull(
            input_energies=self.input_energies,
            energy_key=self.energy_key,
            varying_element=self.intercalating_ion,
            end_members=[self.charged_composition, self.discharged_composition],
            shared_element_basis=self.basis_ion[0],
        )

        d = hull.results
        return d

    @property
    def input_to_voltage_curve(self):
        d = self.mixing_results
        out = {}
        for original_formula in d:
            stability = d[original_formula]["stability"]
            E_per_at = d[original_formula]["E"]
            n_basis_el = CompTools(original_formula).stoich(self.basis_ion[0])
            if n_basis_el == 0:
                continue
            factor = self.basis_ion[1] / n_basis_el
            E_per_basis_fu = E_per_at * CompTools(original_formula).n_atoms * factor

            n_ion = CompTools(original_formula).stoich(self.intercalating_ion)
            n_basis_ion = n_ion * factor

            out[original_formula] = {
                "on_mixing_hull": stability,
                "E_per_at": E_per_at,
                "factor": factor,
                "n_intercalating": n_basis_ion,
                "E": E_per_basis_fu,
            }
        return out

    @property
    def breaks_in_voltage_curve(self):
        d = self.input_to_voltage_curve
        stable_ns = sorted(
            [d[k]["n_intercalating"] for k in d if d[k]["on_mixing_hull"]]
        )
        return stable_ns

    @property
    def plateaus(self):
        breaks = self.breaks_in_voltage_curve
        p = []
        for i, b in enumerate(breaks):
            if i == 0:
                continue
            p.append((breaks[i - 1], b))
        return p

    @property
    def voltages(self):
        z = self.z
        mu_Li = self.mu_Li
        d = self.input_to_voltage_curve
        plateaus = self.plateaus
        voltages = []
        for p in plateaus:
            charged_n, discharged_n = p
            charged_comp = [c for c in d if d[c]["n_intercalating"] == charged_n][0]
            discarged_comp = [c for c in d if d[c]["n_intercalating"] == discharged_n][
                0
            ]
            charged_E = d[charged_comp]["E"]
            discharged_E = d[discarged_comp]["E"]
            delta_n = discharged_n - charged_n

            V = (charged_E + delta_n * mu_Li - discharged_E) / z / delta_n
            voltages.append(V)
        out = {}
        for i, p in enumerate(plateaus):
            out["_".join([str(v) for v in list(p)])] = voltages[i]
        return out


def make_mixing_json():
    fjson = os.path.join("/users/cbartel", "Downloads", "Li2MP2S6_gga_gs_E_per_at.json")
    d = read_json(fjson)

    out = {}
    for M in ["Mn", "Fe", "Co", "Ni"]:
        hull = MixingHull(
            d,
            energy_key="E_per_at",
            varying_element="Li",
            end_members=["Li2%sP2S6" % M, "%sP2S6" % M],
            shared_element_basis=M,
        )
        results = hull.results
        for c in results:
            out[c] = results[c]

    for c in out:
        print("\n")
        print(c)
        print("Emix : %.5f" % out[c]["E_mix"])
        print("x : %.2f" % out[c]["x"])
        print("stability : %s" % out[c]["stability"])

    write_json(out, fjson.replace(".json", "_cjb.json"))
    return out


def make_voltages():
    fjson = os.path.join("/users/cbartel", "Downloads", "Li2MP2S6_gga_gs_E_per_at.json")
    d = read_json(fjson)
    out = {}
    for M in ["Mn", "Fe", "Co", "Ni"]:
        vc = VoltageCurve(
            input_energies=d,
            energy_key="E_per_at",
            intercalating_ion="Li",
            charged_composition="%sP2S6" % M,
            discharged_composition="Li2%sP2S6" % M,
            basis_ion=(M, 1),
        )
        out[M] = vc.voltages

    for M in out:
        print("\n%s" % M)
        print(out[M])
    write_json(out, fjson.replace(".json", "_voltages_cjb.json"))
    return out


def chrisc_main():
    fjson = os.path.join("/users/cbartel", "Downloads", "Li2MP2S6_gga_gs_E_per_at.json")
    energies = read_json(fjson)
    mixing = make_mixing_json()
    voltages = make_voltages()
    return energies, mixing, voltages


def mm_stuff():
    print(os.getcwd())

    fjson = "../data/input_energies.json"
    energies = read_json(fjson)

    energies["Ba1S3Zr1"] = {"E_per_at": -100}
    out = {}
    for M in ["Nb", "Ta"]:
        mixing = MixingHull(
            energies,
            energy_key="E_per_at",
            varying_element=M,
            end_members=["BaZrS3", "Ba%sS3" % M],
            shared_element_basis="Ba",
        )

        print(mixing.relevant_compounds)
        out[M] = mixing.results

    return energies, out


def main():
    fjson = os.path.join("/Users/cbartel/Downloads/gs_ABX3-subs_b.json")
    gs = read_json(fjson)
    input_energies = gs["dmc"]["gga"]
    input_energies["Ba1S3Zr1"] = {"E": -6.391578691}

    input_energies = {
        c: {"E": input_energies[c]["E"]}
        for c in input_energies
        if input_energies[c]["E"]
    }
    out = {}
    for M in ["Nb", "Ta"]:
        mixing = MixingHull(
            input_energies,
            energy_key="E",
            varying_element=M,
            end_members=["BaZrS3", "Ba%sS3" % M],
            shared_element_basis="Ba",
        )

        print(mixing.relevant_compounds)
        out[M] = mixing.results

    results = out["Nb"]
    x = 0.0625
    compound_at_x = [c for c in results if results[c]["x"] == x][0]
    E_mix = results[compound_at_x]["E_mix"]
    print(E_mix)
    return out


if __name__ == "__main__":
    out = main()



from pymatgen.analysis.phase_diagram import PDEntry, PhaseDiagram, CompoundPhaseDiagram
from pymatgen.core.composition import Composition
from pymatgen.analysis.interface_reactions import InterfacialReactivity

from pydmclab.utils.handy import read_json, write_json


class MixingHull(object):
    def __init__(self, input_energies, end_members, energy_key):
        for c in input_energies:
            input_energies[c]["comp"] = Composition(c)
            input_energies[c]["entry"] = PDEntry(
                input_energies[c]["comp"], input_energies[c][energy_key]
            )

        self.input_energies = input_energies
        self.end_members = [Composition(c) for c in end_members]

    @property
    def pd(self):
        input_energies = self.input_energies
        pd_entries = [input_energies[c]["entry"] for c in input_energies]
        return CompoundPhaseDiagram(pd_entries, self.end_members, False)


if __name__ == "__main__":
    mix = main()



from pydmclab.core.comp import CompTools
from pydmclab.core.query import MPQuery
from pydmclab.core.hulls import GetHullInputData, AnalyzeHull, ParallelHulls, MixingHull
from pydmclab.utils.handy import read_json, write_json
from pydmclab.plotting.utils import set_rc_params, get_colors
import matplotlib.pyplot as plt

import os
import numpy as np

set_rc_params()

""" 
Purpose: 
    The convex hull analysis is central to understanding the stability of materials
        Convex hulls are made from formation energy vs composition analyses for a given chemical space
          Standard hulls using formation energies, which are referenced to elemental phases
            These formation energies can be DFT dEf(0 K), dGf(T, p), dphi_f(T, p, mu) etc
          Hulls can also be made from an arbitrary reference (e.g., energies relative to chosen end-member compounds)
            In pydmclab, these are called "Mixing Hulls"
    
    Typical hulls (elemental end-members, n-dimensional for n elements in a chemical space)
        Serial mode:
          - GetHullInputData prepares the input
          - AnalyzeHull processes it
        Parallel mode:
          - ParallelHulls does it all
    
    Mixing hulls (compound end-members, 2D supported, n-d coming later)
        MixingHulls
  
Unit tests:
  - see pydmclab/tests/test_hulls.py

"""
# Chris B's API key for MP query
API_KEY = "N3KdATtMmcsUL94g"

# chemical system to test on
CHEMSYS = "Li-Nb-O-F"

# where to save data
DATA_DIR = os.path.join("output", "hulls", "data")
if not os.path.exists(DATA_DIR):
    os.mkdir(DATA_DIR)

# where to save figures
FIG_DIR = os.path.join("output", "hulls", "figures")
if not os.path.exists(FIG_DIR):
    os.mkdir(FIG_DIR)


def get_mp_data_for_chemsys(
    comp=CHEMSYS,
    only_gs=True,
    data_dir=DATA_DIR,
    remake=False,
):
    """
    Query MP to get some data for a chemical system to paly around with

    Args:
        chemsys (str)
            chemical system to query ('-'.join([elements]))

        only_gs (bool):
            if True, remove non-ground state polymorphs from MP query
                good practice to do this before doing hull analysis b/c non-gs polymorphs trivially have Ehull=Ehull_gs + dE_polymorph-gs

        data_dir (str):
            directory to save data

        remake (bool):
            if True, re-query MP

    Returns:
        gs (dict)
            dictionary of ground state data from MPQuery for chemsys
                {formula (str) : {'Ef_mp' : formation energy (eV/atom)}
    """
    fjson = os.path.join(data_dir, "query_" + comp + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    mpq = MPQuery(API_KEY)
    d = mpq.get_data_for_comp(comp=comp, only_gs=only_gs)

    # data is MPID-keyed, let's make it formula-keyed
    out = {}
    formulas = list(set([d[k]["cmpd"] for k in d]))
    for formula in formulas:
        # NOTE: this only works because I know my query only has 1 structure per formula (only_gs = True)
        gs_key = [k for k in d if d[k]["cmpd"] == formula][0]
        out[formula] = d[gs_key]
    return write_json(out, fjson)


def serial_get_hull_input_data(
    gs, formation_energy_key="Ef_mp", remake=False, data_dir=DATA_DIR, chemsys=CHEMSYS
):
    """
    Let's prepare the hull input file for a "typical" (element-bounded) nD hull

    Args:
        gs (dict)
            dictionary of ground state data from MPQuery for chemsys
                {formula (str) : {'Ef_mp' : formation energy (eV/atom)}

        formation_energy_key (str):
            key to use for formation energy in gs
                'Ef_mp' is default behavior in MPQuery

        remake (bool)
            if True, re-calculate hull input data

        data_dir (str)
            directory to save data

        chemsys (str)
            chemical system to query ('-'.join([elements]))
                just used to generate the savename for the .json

    Returns:
        hullin (dict):
            dictionary of hull input data for gs
                {chemical space (str) :
                    {formula (str) :
                        {'E' : formation energy (float),
                         'amts' :
                            {el (str) :
                                fractional amt of el in formula (float) for el in space
                                }
                            }
                                for all relevant formulas including elements}
                    - elements are automatically given formation energy = 0
                    - chemical space is now in 'el1_el2_...' format to be jsonable
                    - each "chemical space" is a convex hull that must be computed
    """
    fjson = os.path.join(data_dir, "hullin_serial_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    ghid = GetHullInputData(gs, formation_energy_key=formation_energy_key)
    return ghid.hullin_data(fjson=fjson, remake=remake)


def serial_get_hull_output_data(
    hullin, remake=False, chemsys=CHEMSYS, data_dir=DATA_DIR
):
    """
    Now, let's solve the hull we initialized

    Args:
        hullin (dict)
            dictionary of hull input data for chemical system

        remake (bool)
            if True, re-calculate hull output data

        chemsys (str)
            chemical system to query ('-'.join([elements]))
                only used for savename

        data_dir (str)
            directory to save data

    Returns:
        stability data (dict) for all compounds in the specified chemical space
            {compound (str) :
                {'Ef' : formation energy (float),
                 'Ed' : decomposition energy (float),
                 'rxn' : decomposition reaction (str),
                 'stability' : stable (True) or unstable (False)
                 }
                 for all compounds in the chemical space
            }
    """
    fjson = os.path.join(data_dir, "hullout_serial_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    hullout = {}
    # loop through each chemical (sub)space in the hullin dictionary
    for space in hullin:
        ah = AnalyzeHull(hullin, space)
        # loop through every compound in that space
        for cmpd in hullin[space]:
            print("\n%s" % cmpd)
            # get the stability data for that compound
            hullout[cmpd] = ah.cmpd_hull_output_data(cmpd)
    return write_json(hullout, fjson)


def parallel_get_hull_input_and_output_data(
    gs, remake=False, chemsys=CHEMSYS, data_dir=DATA_DIR, n_procs=2, fresh_restart=True
):
    """
    Repeat what we did with GetHullInputData + AnalyzeHull, but now in parallel

    Args:
        gs (dict)
            dictionary of ground state data from MPQuery for chemsys
                {formula (str) : {'Ef_mp' : formation energy (eV/atom)}}
                }
        remake (bool)
            if True, re-calculate hull input and output data

        chemsys (str)
            chemical system to query ('-'.join([elements]))
                only for savename

        data_dir (str)
            directory to save data

        n_procs (int)
            number of processors to use
                could also be 'all' to use multip.cpu_count()-1 procs

        fresh_restart (bool)
            if True, restart ParallelHull process from scratch
            if False, use the files we've already written from a previous execution

    Returns:
        stability data (dict) for all compounds in the specified chemical space
            {compound (str) :
                {'Ef' : formation energy (float),
                 'Ed' : decomposition energy (float),
                 'rxn' : decomposition reaction (str),
                 'stability' : stable (True) or unstable (False)
                 }
                 for all compounds in the chemical space
            }
    """
    fjson = os.path.join(data_dir, "hullout_parallel_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    ph = ParallelHulls(gs, n_procs=n_procs, fresh_restart=fresh_restart)

    # prepare hull input data
    hullin = ph.parallel_hullin(fjson=fjson.replace("hullout", "hullin"))

    # identify the simplest hull that can be solved for each compound
    smallest_spaces = ph.smallest_spaces(
        hullin=hullin, fjson=fjson.replace("hullout", "small_spaces")
    )

    # solve these minimally complex hulls for every compound
    return ph.parallel_hullout(
        hullin=hullin, smallest_spaces=smallest_spaces, fjson=fjson, remake=True
    )


def plot_to_check_success(gs, serial_hullout, parallel_hullout):
    """
    Let's make sure things worked
        1) Parallel and serial should give the same results
        2) For unstable compounds (above the hull), our Ed results and MP E above hull should be the same
        3) For stable compounds, our Ed results will be negative; MP's should be E above hull = 0

    Args:
        gs (dict)
            dictionary of ground state data from MPQuery for chemsys

        serial_hullout (dict)
            dictionary of hull output data for gs (run serially)

        parallel_hullout (dict)
            dictionary of hull output data for gs (run in parallel)

    Returns:
        a couple scatter plots

    """
    set_rc_params()

    fig = plt.figure(figsize=(8, 3))

    params = {
        "serial": {"m": "o", "c": "blue"},
        "parallel": {"m": "^", "c": "orange"},
    }

    cmpds = sorted(gs.keys())

    cmpds = [c for c in cmpds if CompTools(c).n_els > 1]

    mp_Ehull = [gs[c]["Ehull_mp"] for c in cmpds]

    serial_decomp = [serial_hullout[c]["Ed"] for c in cmpds]
    parallel_decomp = [parallel_hullout[c]["Ed"] for c in cmpds]

    x = mp_Ehull
    y1 = serial_decomp
    y2 = parallel_decomp

    ax1 = plt.subplot(121)

    ax1 = plt.scatter(y2, y1, edgecolor="blue", marker="o", color="white")

    # ax1 = plt.xticks(xticks[1])
    # ax1 = plt.yticks(yticks[1])
    xlim, ylim = (-0.5, 1), (-0.5, 1)
    ax1 = plt.xlabel("Ed from parallel (eV/at)")
    ax1 = plt.ylabel("Ed from serial (eV/at)")
    ax1 = plt.plot(xlim, xlim, color="black", lw=1, ls="--")
    ax1 = plt.xlim(xlim)
    ax1 = plt.ylim(ylim)

    ax2 = plt.subplot(122)
    ax2 = plt.scatter(x, y1, edgecolor="blue", marker="o", color="white")

    # ax1 = plt.xticks(xticks[1])
    # ax1 = plt.yticks(yticks[1])
    xlim, ylim = (-0.1, 1), (-1, 1)
    ax2 = plt.xlabel("Ehull from MP (eV/at)")
    ax2 = plt.ylabel("")
    ax2 = plt.plot(xlim, xlim, color="black", lw=1, ls="--")
    ax2 = plt.gca().yaxis.set_ticklabels([])
    ax2 = plt.xlim(xlim)
    ax2 = plt.ylim(ylim)

    disagreements = []
    for k in serial_hullout:
        if CompTools(k).n_els == 1:
            continue
        if serial_hullout[k]["stability"] and (gs[k]["Ehull_mp"] > 0):
            disagreements.append(k)
        if not serial_hullout[k]["stability"] and (gs[k]["Ehull_mp"] == 0):
            disagreements.append(k)

        if (gs[k]["Ehull_mp"] != 0) and (
            np.round(serial_hullout[k]["Ed"], 3) != np.round(gs[k]["Ehull_mp"], 3)
        ):
            disagreements.append(k)

    for k in disagreements:
        print("\n%s" % k)
        print("my rxn = %s" % serial_hullout[k]["rxn"])
        print("my hull = %.3f" % serial_hullout[k]["Ed"])
        print("mp hull = %.3f" % gs[k]["Ehull_mp"])

    # plt.show()

    fig.savefig(os.path.join(FIG_DIR, "pd_demo_check.png"))


def demo_mixing_hull(gs, end_members, remake=False):
    """
    Now, we'll do a "mixing hull"
        A convex hull with compounds as end members

    e.g., a 2D hull bounded by LiNbO3-Li2NbO3

    Args:
        gs (dict)
            dictionary of ground state data from MPQuery for chemsys

        end_members (list)
            list of end members to use for mixing hull

        remake (bool)

    """
    fjson = os.path.join(
        DATA_DIR, "pd_demo_mixing_hull_%s-%s.json" % (end_members[0], end_members[1])
    )
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    mh = MixingHull(
        input_energies=gs,
        varying_element="Li",
        end_members=end_members,
        shared_element_basis="Nb",
        energy_key="Ef_mp",
    )
    out = mh.results
    for k in out:
        if out[k]["stability"]:
            print(k)
            print(out[k]["x"])


def main():
    # if True, re-grab data from MP
    remake_query = False
    # if True, re-calculate hull input data
    remake_serial_hullin = False
    # if True, re-calculate hull output data
    remake_serial_hullout = False
    # if True, re-calculate hull output data in parallel
    remake_parallel_hullout = False
    # if True, generate figure to check results
    remake_hull_figure_check = False
    # if True, remake mixing hull
    remake_mixing_hull = False

    # MP query for CHEMSYS
    gs = get_mp_data_for_chemsys(CHEMSYS, remake=remake_query)

    # hull input data for CHEMSYS
    hullin = serial_get_hull_input_data(gs, remake=remake_serial_hullin)

    # hull output data for CHEMSYS
    hullout = serial_get_hull_output_data(hullin, remake=remake_serial_hullout)

    # hull output data for CHEMSYS (generated using parallelization)
    p_hullout = parallel_get_hull_input_and_output_data(
        gs, remake=remake_parallel_hullout
    )

    # generate a graph that compares serial vs parallel hull output and also compares ThermoTools hull output to MP hull data
    if remake_hull_figure_check:
        plot_to_check_success(gs, hullout, p_hullout)

    # generate a mixing hull
    mixing = demo_mixing_hull(gs, ["NbO2", "LiNbO2"], remake=remake_mixing_hull)
    return gs, hullin, hullout, p_hullout


if __name__ == "__main__":
    # MP Query --> hull input data (serial) --> hull output data (serial) --> hull output data (parallel)
    gs = main()



from pydmc.core.mag import MagTools

from pydmclab.core.query import MPQuery

""" 
Purpose: 
  MagTools is used to generate magnetic orderings for structures
  
Unit tests:
  - see pydmclab/tests/test_mag.py

"""


def demo_get_afm_structures():
    """
    gets AFM orderings for Cr2O3
    """

    comps = ["Cr2O3"]
    mpq = MPQuery(api_key="N3KdATtMmcsUL94g")
    q = mpq.get_data_for_comp(comps, only_gs=True)
    q = {q[k]["cmpd"]: q[k] for k in q}

    Cr2O3 = q["Cr2O3"]["structure"]

    mt = MagTools(Cr2O3)

    strucs = mt.get_antiferromagnetic_structures

    print(strucs[1])


def main():
    demo_get_afm_structures()
    return


if __name__ == "__main__":
    main()



from pydmclab.utils.handy import read_json
import matplotlib.pyplot as plt
import os
from pydmclab.plotting.dos import ax_tdos
from pydmclab.plotting.utils import get_colors


def main():
    test_data_dir = "../data/test_data/vasp/AlN"
    tdos = read_json(os.path.join(test_data_dir, "tdos.json"))

    fig = plt.figure()
    ax1 = plt.subplot(121)
    ax1 = ax_tdos(
        tdos,
        population_sources=["total", "Al"],
        colors={"total": "black", "Al": "green"},
        color_palette=get_colors("set2"),
        params={"line_alpha": 0.9, "fill_alpha": 0.2, "lw": 1},
        special_labels=None,
        spins="summed",
        normalization=1,
        smearing=1,
        Efermi=0.0,
        xlim=(0, 10),
        ylim=(-2, 2),
        xticks=(True, [0, 5, 10]),
        yticks=(True, [-2, -1, 0, 1, 2]),
        xlabel="DOS",
        ylabel=r"$E-E_F\/(eV)$",
        legend=True,
        title=None,
        savename=None,
        show=False,
    )

    ax2 = plt.subplot(122)
    ax2 = ax_tdos(
        tdos,
        population_sources=["total", "Al"],
        colors={"total": "black", "Al": "orange"},
        color_palette=get_colors("set2"),
        params={"line_alpha": 0.9, "fill_alpha": 0.2, "lw": 1},
        special_labels=None,
        spins="summed",
        normalization=1,
        smearing=2,
        Efermi=0.0,
        xlim=(0, 10),
        ylim=(-2, 2),
        xticks=(True, [0, 5, 10]),
        yticks=(False, [-2, -1, 0, 1, 2]),
        xlabel="DOS",
        ylabel="",
        legend=True,
        title=None,
        savename=None,
        show=False,
    )

    return tdos


if __name__ == "__main__":
    tdos = main()



from pydmclab.core.comp import CompTools
from pydmclab.core.hulls import GetHullInputData, AnalyzeHull, MixingHull
from pydmclab.plotting.utils import set_rc_params, get_colors
from pydmclab.utils.handy import read_json, write_json

import os
import numpy as np

import matplotlib.pyplot as plt
import matplotlib as mpl

import itertools

data_dir = "../data/test_data/ternarypd"


def get_data():
    return read_json(os.path.join(data_dir, "Mg-Cr-S-Na-Cl_query.json"))


def get_small_data(remake=False):
    fjson = os.path.join(data_dir, "Efs.json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    d = get_data()
    out = {k: {"Ef": d[k]["Ef_mp"]} for k in d}

    write_json(out, fjson)
    return read_json(fjson)


def get_hullout(Efs, remake=False):
    fjson = os.path.join(data_dir, "hullout.json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    ghid = GetHullInputData(Efs, "Ef")
    hullin = ghid.hullin_data()
    out = {}
    for space in hullin:
        hullout = AnalyzeHull(hullin, space).hull_output_data
        out.update(hullout)

    write_json(out, fjson)
    return read_json(fjson)


def get_mixing(hullout, remake=False):
    fjson = os.path.join(data_dir, "mixing.json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    hullout = {k: hullout[k] for k in ["Cl1Na1", "Cl2Mg1", "Cl4Mg1Na2", "Cl8Mg1Na6"]}
    print(hullout.keys())
    mh = MixingHull(
        hullout,
        varying_element="Na",
        end_members=["NaCl", "MgCl2"],
        shared_element_basis="Cl",
        energy_key="Ef",
    )

    print(mh.mixing_energies)
    out = mh.results

    write_json(out, fjson)
    return read_json(fjson)


class BinaryPD(object):
    def __init__(
        self,
        stability_data,
        end_members,
        polymorph_data={},
        stability_source="AnalyzeHull",
    ):
        """

        Args:
            stability_data (dict):
                AnalyzeHull(...).hull_output_data
                    {formula (str) :
                        {'Ef' : formation energy (eV/atom),
                        'stability' : True if on the hull else False}
                        }

                OR

                MixingHull(...).results
                    {formula (str) :
                        {'E_mix' : mixing energy (eV/atom),
                         'stability' : True if on the mixing hull else False,
                         'x' : mole fraction of right end member}}}



            end_members (list): [left end, right end]
                e.g., ['Mg', 'S'] for a typical Mg-S phase diagram (AnalyzeHull)
                e.g., ['MgCl2', 'NaCl'] for a mixing hull laong NaCl-MgCl2 (MixingHull)

            polymorph_data (dict):
                {formula (str) :
                    {ID (str) :
                        {'dE_gs' : energy above ground state (eV/atom)}}}

            stability_source (str): 'AnalyzeHull' or 'MixingHull'

        Returns:
            data (dict):
                {formula (str) :
                    {'Ef' : formation energy (eV/atom),
                     'x' : mole fraction of right end member,
                     'stability' : True if on the hull else False,
                     'polys' : [{'ID' : ID (str), 'Ef' : Ef_gs + dE_gs (eV/atom)}]}}


            left_end (str): left end member
            right_end (str): right end member
        """
        data = stability_data.copy()
        left_end, right_end = end_members

        if stability_source == "AnalyzeHull":
            if (CompTools(end_members[0]).n_els != 1) or (
                CompTools(end_members[1]).n_els != 1
            ):
                raise ValueError(
                    "end_members must be elements if source is AnalyzeHull; use MixingHull for other situations"
                )

            els = [CompTools(end_members[0]).els[0], CompTools(end_members[1]).els[0]]
            data = {
                k: data[k] for k in data if set(CompTools(k).els).issubset(set(els))
            }

            for formula in data:
                x = CompTools(formula).mol_frac(right_end)
                data[formula]["x"] = x

        elif stability_source == "MixingHull":
            for formula in data:
                data[formula]["Ef"] = data[formula]["E_mix"]

        for formula in stability_data:
            data[formula]["polys"] = []
            if formula in polymorph_data:
                for ID in polymorph_data[formula]:
                    dE_gs = polymorph_data[formula][ID]["dE_gs"]
                    data[formula]["polys"].append(
                        {"ID": ID, "Ef": data[formula]["Ef"] + dE_gs}
                    )

        self.data = data
        self.left_end, self.right_end = left_end, right_end

    @property
    def sorted_xs(self):
        """

        Returns:
            list of x values sorted from 0 to 1
        """
        data = self.data
        return sorted([data[k]["x"] for k in data])

    @property
    def stable(self):
        """


        Returns:
            list of dicts
                {'x' : mole fraction of right end member,
                 'Ef' : formation energy (eV/atom),
                 'label' : formula}
            sorted by x (ascending)
            only for stable compounds
        """
        data = self.data
        data = {k: data[k] for k in data if data[k]["stability"]}
        xs = self.sorted_xs
        to_plot = []
        for x in xs:
            formula = [k for k in data if data[k]["x"] == x]
            if formula:
                formula = formula[0]
            else:
                continue
            to_plot.append({"x": x, "Ef": data[formula]["Ef"], "label": formula})

        return to_plot

    @property
    def unstable(self):
        """
        Returns:
            list of dicts
                {'x' : mole fraction of right end member,
                 'Ef' : formation energy (eV/atom),
                 'label' : formula (for gs) or ID (for non-gs)}
            sorted by x (ascending)
            only for unstable compounds
        """
        data = self.data
        to_plot = []
        for formula in data:
            polys = data[formula]["polys"]
            x = data[formula]["x"]
            for p in polys:
                to_plot.append({"x": x, "Ef": p["Ef"], "label": p["ID"]})
            if not data[formula]["stability"]:
                to_plot.append({"x": x, "Ef": data[formula]["Ef"], "label": formula})

        return to_plot

    def ax_pd(
        self,
        color_palette=get_colors("tab10"),
        stable_params={
            "markerfacecolor": "white",
            "color": "blue",
            "marker": "o",
            "markersize": 8,
            "lw": 2,
        },
        unstable_params={
            "color": "white",
            "edgecolor": "red",
            "marker": "^",
            "s": 80,
        },
        label_ends=True,
        el_order=None,
        xlim=(0, 1),
        ylim=(-5, 1),
        xticks=(True, [0, 0.2, 0.4, 0.6, 0.8, 1]),
        yticks=(True, [-5, -4, -3, -2, -1, 0, 1]),
        xlabel="auto",
        ylabel=r"$\Delta$" + r"$\it{E}$" + r"$_f$" + r"$\/(\frac{eV}{atom})$",
        legend=True,
        title=None,
        savename=None,
        show=True,
    ):
        """_summary_

        Args:
            color_palette (_type_, optional): _description_. Defaults to get_colors("tab10").
            stable_params (dict, optional): _description_. Defaults to { "markerfacecolor": "white", "color": "blue", "marker": "o", "markersize": 8, "lw": 2, }.
            unstable_params (dict, optional): _description_. Defaults to { "color": "white", "edgecolor": "red", "marker": "^", "s": 80, }.
            label_ends (bool, optional): _description_. Defaults to True.
            el_order (_type_, optional): _description_. Defaults to None.
            xlim (tuple, optional): _description_. Defaults to (0, 1).
            ylim (tuple, optional): _description_. Defaults to (-5, 1).
            xticks (tuple, optional): _description_. Defaults to (True, [0, 0.2, 0.4, 0.6, 0.8, 1]).
            yticks (tuple, optional): _description_. Defaults to (True, [-5, -4, -3, -2, -1, 0, 1]).
            xlabel (str, optional): _description_. Defaults to "auto".
            ylabel (regexp, optional): _description_. Defaults to r"$\Delta$"+r"$\it{E}$"+r"$"+r"$\/(\frac{eV}{atom})$".
            legend (bool, optional): _description_. Defaults to True.
            title (_type_, optional): _description_. Defaults to None.
            savename (_type_, optional): _description_. Defaults to None.
            show (bool, optional): _description_. Defaults to True.
        """
        stable_params["color"] = color_palette[stable_params["color"]]
        unstable_params["edgecolor"] = color_palette[unstable_params["edgecolor"]]

        stable_to_plot = self.stable
        unstable_to_plot = self.unstable

        left_end, right_end = self.left_end, self.right_end

        x, y, labels = [], [], []
        for d in stable_to_plot:
            x.append(d["x"])
            y.append(d["Ef"])
            labels.append(d["label"])

        ax = plt.plot(x, y, label="stable", **stable_params, zorder=1)

        x, y, labels = [], [], []
        for d in unstable_to_plot:
            x.append(d["x"])
            y.append(d["Ef"])
            labels.append(d["label"])

        ax = plt.scatter(x, y, label="unstable", **unstable_params, zorder=0)

        ax = plt.xticks(visible=xticks[0], ticks=xticks[1])
        ax = plt.yticks(visible=yticks[0], ticks=yticks[1])
        ax = plt.xlim(xlim)
        ax = plt.ylim(ylim)

        if label_ends:
            xpos, ypos = 0, 1.1 * max(yticks[1])
            ax = plt.text(
                xpos,
                ypos,
                get_label(left_end, el_order),
                fontsize=20,
                ha="center",
                va="bottom",
            )
            xpos, ypos = 1, 1.1 * max(yticks[1])
            ax = plt.text(
                xpos,
                ypos,
                get_label(right_end, el_order),
                fontsize=20,
                ha="center",
                va="bottom",
            )
        if xlabel == "auto":
            xlabel = (
                r"$x\/in\/$"
                + "(%s)" % get_label(left_end, el_order)
                + r"$_{1-x}$"
                + "(%s)" % get_label(right_end, el_order)
                + r"$_x$"
            )
        ax = plt.xlabel(xlabel)
        ax = plt.ylabel(ylabel)

        if legend:
            ax = plt.legend(loc=legend if isinstance(legend, str) else "best")


class TernaryPD(object):
    def __init__(self, stability_data, end_members):
        """

        Args:
            stability_data (dict):
                AnalyzeHull(...).hull_output_data
                    {formula (str) :
                        {'Ef' : formation energy (eV/atom),
                        'stability' : True if on the hull else False}
                        }

                NOTE: MixingHull results not supported (yet) for ternary diagrams

            end_members (list):
                list of end members (str) in the ternary system [right, top, left]
                NOTE: only elemental end members supported (for now)

        Returns:
            data (dict):
                {formula (str) :
                    {'Ef' : formation energy (eV/atom),
                     'stability' : True if on the hull else False,
                     'a' : mol fraction of right end member,
                     'b' : mol fraction of top end member,
                     'c' : mol fraction of left end member,
                     'x' : x coordinate of formula in ternary diagram,
                     'y' : y coordinate of formula in ternary diagram,}

            left_end (str):
                left end member (str) in the ternary system

            right_end (str):
                right end member (str) in the ternary system

            top_end (str):
                top end member (str) in the ternary system
        """
        stability_data = stability_data.copy()

        for e in end_members:
            if CompTools(e).n_els != 1:
                raise ValueError("Only elemental end members supported (for now)")
        right_end, top_end, left_end = [CompTools(e).els[0] for e in end_members]
        els_in_end_members = [right_end, top_end, left_end]

        data = {
            k: stability_data[k]
            for k in stability_data
            if set(CompTools(k).els).issubset(set(els_in_end_members))
        }
        for formula in data:
            a, b, c = [CompTools(formula).mol_frac(el) for el in els_in_end_members]
            data[formula].update({"a": a, "b": b, "c": c})
            x, y = triangle_to_square((a, b, c))
            data[formula].update({"x": x, "y": y})

        self.data = data
        self.right_end, self.top_end, self.left_end = els_in_end_members

    @property
    def points_to_plot(self):
        """

        Returns:
            list of dicts
                {'x' : x coordinate of formula in ternary diagram,
                 'y' : y coordinate of formula in ternary diagram,
                 'Ef' : formation energy (eV/atom),
                 'stability' : True if on the hull else False,
                 'formula' : formula (str)}
        """
        data = self.data
        to_plot = []
        for formula in data:
            to_plot.append(
                {
                    "x": data[formula]["x"],
                    "y": data[formula]["y"],
                    "Ef": data[formula]["Ef"],
                    "stability": data[formula]["stability"],
                    "formula": formula,
                }
            )
        return to_plot

    @property
    def space(self):
        """_summary_

        Returns:
            space (str) of the ternary diagram right-el_top-el_left-el
        """
        return "_".join(sorted([self.right_end, self.top_end, self.left_end]))

    @property
    def hullin(self):
        """_summary_

        Returns:
            {formula (str) :
                {'E' : formation energy (eV/atom),
                 'amts' : {el (str) : mol fraction of el in space (float)}}
        """
        data = self.data
        space = self.space
        out = {
            space: {
                c: {
                    "E": data[c]["Ef"],
                    "amts": {el: CompTools(c).mol_frac(el) for el in space.split("_")},
                }
                for c in data
            }
        }
        return out

    @property
    def lines_to_plot(self):
        """_summary_

        Returns:
            list of dicts
                {'x' : x endpoints of line in ternary diagram,
                 'y' : y endpoints of line in ternary diagram}
        """
        data = self.data
        hullin = self.hullin
        ah = AnalyzeHull(hullin, self.space)
        simplices = ah.hull_simplices
        lines = uniquelines(simplices)
        sorted_compounds = ah.sorted_compounds
        to_plot = []
        for l in lines:
            compounds = (sorted_compounds[l[0]], sorted_compounds[l[1]])

            a0, b0, c0 = [data[compounds[0]][v] for v in ["a", "b", "c"]]
            a1, b1, c1 = [data[compounds[1]][v] for v in ["a", "b", "c"]]
            x0, y0 = triangle_to_square((a0, b0, c0))
            x1, y1 = triangle_to_square((a1, b1, c1))
            to_plot.append({"x": [x0, x1], "y": [y0, y1]})

        return to_plot

    def ax_pd(
        self,
        color_palette=get_colors("tab10"),
        stable_params={
            "color": "white",
            "edgecolor": "blue",
            "marker": "o",
            "s": 50,
        },
        unstable_params={
            "color": "white",
            "edgecolor": "red",
            "marker": "^",
            "s": 50,
        },
        label_els=True,
        label_compounds=[],
        legend=True,
        title=None,
        savename=None,
        show=True,
    ):
        """_summary_

        Args:
            color_palette (_type_, optional): _description_. Defaults to get_colors("tab10").
            stable_params (dict, optional): _description_. Defaults to { "color": "white", "edgecolor": "blue", "marker": "o", "s": 50, }.
            unstable_params (dict, optional): _description_. Defaults to { "color": "white", "edgecolor": "red", "marker": "^", "s": 50, }.
            label_els (bool, optional): _description_. Defaults to True.
            label_compounds (list, optional): _description_. Defaults to [].
            legend (bool, optional): _description_. Defaults to True.
            title (_type_, optional): _description_. Defaults to None.
            savename (_type_, optional): _description_. Defaults to None.
            show (bool, optional): _description_. Defaults to True.
        """
        data = self.data

        stable_params["edgecolor"] = color_palette[stable_params["edgecolor"]]
        unstable_params["edgecolor"] = color_palette[unstable_params["edgecolor"]]

        points_to_plot = self.points_to_plot
        stable_to_plot = [d for d in points_to_plot if d["stability"]]
        unstable_to_plot = [d for d in points_to_plot if not d["stability"]]

        x, y, labels = [], [], []
        for d in stable_to_plot:
            x.append(d["x"])
            y.append(d["y"])
            labels.append(d["formula"])

        ax = plt.scatter(x, y, label="stable", **stable_params, zorder=2)

        x, y, labels = [], [], []
        for d in unstable_to_plot:
            x.append(d["x"])
            y.append(d["y"])
            labels.append(d["formula"])

        ax = plt.scatter(x, y, label="unstable", **unstable_params, zorder=1)

        lines_to_plot = self.lines_to_plot
        for l in lines_to_plot:
            ax = plt.plot(l["x"], l["y"], color="black", zorder=0)

        for spine in ["bottom", "left", "top", "right"]:
            ax = plt.gca().spines[spine].set_visible(False)
        ax = plt.gca().xaxis.set_ticklabels([])
        ax = plt.gca().yaxis.set_ticklabels([])
        ax = plt.gca().tick_params(bottom=False, top=False, left=False, right=False)

        if label_els:
            el_label_size = 18
            left_el_pos = (-0.02, -0.05)
            right_el_pos = (1.02, -0.05)
            top_el_pos = (0.5, 0.89)
            ax = plt.text(
                right_el_pos[0],
                right_el_pos[1],
                self.right_end,
                fontsize=el_label_size,
                horizontalalignment="left",
            )
            ax = plt.text(
                top_el_pos[0],
                top_el_pos[1],
                self.top_end,
                fontsize=el_label_size,
                horizontalalignment="center",
            )
            ax = plt.text(
                left_el_pos[0],
                left_el_pos[1],
                self.left_end,
                fontsize=el_label_size,
                horizontalalignment="right",
            )

        if legend:
            ax = plt.legend(loc=legend if isinstance(legend, str) else "best")


def triangle_to_square(pt):
    """
    Args:
        pt (tuple) - (left, top, right) triangle point
                e.g., (1, 0, 0) is left corner
                e.g., (0, 1, 0) is top corner

    Returns:
        pt (tuple) - (x, y) same point in 2D space
    """
    a, b, c = pt
    converter = np.array([[1, 0], [0.5, np.sqrt(3) / 2]])
    new_pt = np.dot(np.array([a, b]), converter)
    return new_pt.transpose()


def square_to_triangle(pt):
    """
    Args:
        pt (tuple) - (x, y) planar point in 2D space

    Returns:
        pt (tuple) - (left, top, right) triangle point
                e.g., (1, 0, 0) is left corner
                e.g., (0, 1, 0) is top corner
    """
    x, y = pt
    b = y * 2 / np.sqrt(3)
    a = x - 0.5 * b
    c = 1 - a - b
    return (a, b, c)


def uniquelines(q):
    """
    Given all the facets, convert it into a set of unique lines.  Specifically
    used for converting convex hull facets into line pairs of coordinates.

    Args:
        q: A 2-dim sequence, where each row represents a facet. E.g.,
            [[1,2,3],[3,6,7],...]

    Returns:
        setoflines:
            A set of tuple of lines.  E.g., ((1,2), (1,3), (2,3), ....)
    """
    setoflines = set()
    for facets in q:
        for line in itertools.combinations(facets, 2):
            setoflines.add(tuple(sorted(line)))
    return setoflines


def get_label(cmpd, els=None):
    """
    Args:
        cmpd (str) - chemical formula
        els (list) - ordered list of elements (str) as you want them to appear in label

    Returns:
        neatly formatted chemical formula label
    """
    if not els:
        els = CompTools(cmpd).els
    label = r"$"
    for el in els:
        amt = CompTools(cmpd).stoich(el)
        if amt == 0:
            continue
        label += el
        if amt == 1:
            continue
        label += "_{%s}" % amt
    label += "$"
    return label


def main():
    set_rc_params()
    query = get_small_data()
    hullout = get_hullout(query)
    mixing = get_mixing(hullout, remake=False)
    # bpd_ah = BinaryPD(hullout, ["Mg", "Cl"], stability_source="AnalyzeHull")
    # bpd_ah.ax_pd()
    bpd_mh = BinaryPD(
        mixing,
        ["NaCl", "MgCl2"],
        polymorph_data={
            "Cl4Mg1Na2": {"mp-1234": {"dE_gs": 0.05}},
            "Cl8Mg1Na6": {
                "mp-4321": {"dE_gs": 0.1},
            },
        },
        stability_source="MixingHull",
    )

    # tpd = None

    bpd_mh.ax_pd(
        yticks=(True, [-0.1, 0.1]), ylim=(-0.1, 0.1), el_order=("Na", "Mg", "Cl")
    )

    # tpd = TernaryPD(hullout, ["S", "Cr", "Mg"])
    # tpd.ax_pd()
    tpd = None
    return query, hullout, mixing, bpd, tpd


if __name__ == "__main__":
    query, hullout, mixing, bpd, tpd = main()



from pydmclab.core.comp import CompTools

from pymatgen.core.structure import Structure
from pymatgen.transformations.standard_transformations import (
    OrderDisorderedStructureTransformation,
    AutoOxiStateDecorationTransformation,
    OxidationStateDecorationTransformation,
)
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

import os
import numpy as np


class StrucTools(object):
    """
    Purpose: to manipulate crystal structures for DFT calculations
    """

    def __init__(self, structure, ox_states=None):
        """
        Args:
            structure (Structure): pymatgen Structure object
                - if dict, assumes it is Structure.as_dict(); converts to Structure object
                - if str, assumes it is a path to a structure file, converts to Structure object
            ox_states (dict): dictionary of oxidation states {el (str) : oxidation state (int)}
                - or None

        """
        # convert Structure.as_dict() to Structure
        if isinstance(structure, dict):
            structure = Structure.from_dict(structure)

        # convert file into Structure
        if isinstance(structure, str):
            if os.path.exists(structure):
                structure = Structure.from_file(structure)
            else:
                raise ValueError(
                    "you passed a string to StrucTools > this means a path to a structure > but the path is empty ..."
                )
        self.structure = structure
        self.ox_states = ox_states

    @property
    def structure_as_dict(self):
        """

        Returns:
            dict: pymatgen Structure.as_dict()
        """
        return self.structure.as_dict()

    @property
    def compact_formula(self):
        """
        "clean" (reduced, systematic) formula (str) for structure
        """
        return CompTools(self.structure.formula).clean

    @property
    def formula(self):
        """
        pretty (unreduced formula) for structure
        """
        return self.structure.formula

    @property
    def els(self):
        """
        list of unique elements (str) in structure
        """
        return CompTools(self.compact_formula).els

    @property
    def amts(self):
        """
        Returns:
            {el (str): number of el in struc (int)}
        """
        els = self.els
        amts = {el: 0 for el in els}
        structure = self.structure
        for i, site in enumerate(structure):
            el = SiteTools(structure, i).el
            if el:
                amts[el] += 1
        return amts

    def make_supercell(self, grid):
        """
        Args:
            grid (list) - [nx, ny, nz]

        Returns:
            Structure repeated nx, ny, nz

            so to make a 1x2x3 supercell of the initial structure, use:
                supercell = StrucTools(structure).make_supercell([1, 2, 3])
        """
        structure = self.structure
        print("making supercell with grid %s\n" % str(grid))
        structure.make_supercell(grid)
        return structure

    def perturb(self, perturbation=0.1):
        """
        Args:
            perturbation (float) - distance in Angstrom to randomly perturb each atom

        Returns:
            Structure w/ perturbations
        """
        structure = self.structure
        structure.perturb(perturbation)
        return structure

    def change_occ_for_site(self, site_idx, new_occ, structure=None):
        """

        return a structure with a new occupation for some site

        Args:
            site_idx (int):
                index of site in structure to change

            new_occ (dict):
                dictionary telling me the new occupation on that site
                    e.g., {'Li' : 0.5, 'Fe' : 0.5}

            structure (None or pymatgen Structure object):
                if None, start from self.structure
                else, start from structure
                    (in case you don't want to start from the structure that initialized StrucTools)

        Returns:
            pymatgen Structure object with new occupation
        """

        if not structure:
            structure = self.structure

        s = structure.copy()

        if np.sum(list(new_occ.values())) == 0:
            # if new occupation is 0, remove that site
            s.remove_sites([site_idx])
        else:
            # otherwise, update the occupation
            s[site_idx].species = new_occ
        return s

    def change_occ_for_el(self, el, new_occ, structure=None):
        """
        Args:
            el (str)
                element to change occupation for

            new_occ (dict)
                {el : new occupation (float)}

            structure (None or pymatgen Structure object)
                if None, start from self.structure
        """
        if not structure:
            structure = self.structure

        # for all sites having that element, change the occupation
        for i, site in enumerate(structure):
            if SiteTools(structure, i).el == el:
                structure = self.change_occ_for_site(i, new_occ, structure=structure)

        return structure

    @property
    def decorate_with_ox_states(self):
        """
        Returns oxidation state decorated structure
            - uses Auto algorithm if no ox_states are provided
            - otherwise, applies ox_states
        """
        print("decorating with oxidation states\n")
        structure = self.structure
        ox_states = self.ox_states
        if not ox_states:
            print("     automatically\n")
            transformer = AutoOxiStateDecorationTransformation()
        else:
            transformer = OxidationStateDecorationTransformation(
                oxidation_states=ox_states
            )
            print("     using %s" % str(ox_states))
        return transformer.apply_transformation(structure)

    def get_ordered_structures(self, algo=0, decorate=True, n_strucs=1):
        """
        Args:
            algo (int):
                method for enumeration
                    0 = fast, 1 = complete, 2 = best first
                        see pymatgen.transformations.standard_transformations.OrderDisorderedStructureTransformation
                        0 usually OK

            decorate (bool)
                whether to decorate with oxidation states
                    if False, self.structure must already have them

            n_strucs (int)
                number of ordered structures to return

        Returns:
            dict of ordered structures
            {index : structure (Structure.as_dict())}
                - index = 0 has lowest Ewald energy
        """

        # initialize ordering engine
        transformer = OrderDisorderedStructureTransformation(algo=algo)

        # decorat with oxidation states or not
        if decorate:
            structure = self.decorate_with_ox_states
        else:
            structure = self.structure

        # only return one structure if n_strucs = 1
        return_ranked_list = n_strucs if n_strucs > 1 else False

        # generate ordered structure
        print("ordering disordered structures\n")
        out = transformer.apply_transformation(
            structure, return_ranked_list=return_ranked_list
        )

        if isinstance(out, list):
            # more than 1 structure, so check for duplicates (symmetrically equivalent structures) and remove them
            print("getting unique structures\n")
            matcher = StructureMatcher()
            out = [i["structure"] for i in out]
            # find unique groups of structures
            groups = matcher.group_structures(out)
            out = [groups[i][0] for i in range(len(groups))]
            return {i: out[i].as_dict() for i in range(len(out))}
        else:
            # if only one structure is made, return in same formation (dict)
            return {0: out.as_dict()}

    def replace_species(
        self,
        species_mapping,
        n_strucs=1,
        use_ox_states_in_mapping=False,
        use_occ_in_mapping=True,
    ):
        """
        Args:
            species_mapping (dict)
                {Element(el) :
                    {Element(el1) : fraction el1,
                     Element(el2) : fraction el2,
                     ...},
                 ...}

            n_strucs (int)
                number of ordered structures to return if disordered

            use_ox_states_in_mapping (bool)
                if False, will remove oxidation states before doing replacements

            use_occ_in_mapping (bool)
                if False, will set all occupancies to 1.0 before doing replacements

        Returns:
            dict of ordered structures
                {index : structure (Structure.as_dict())}
                    index = 0 has lowest Ewald energy
        """
        structure = self.structure
        print("replacing species with %s\n" % str(species_mapping))

        # purge oxidation states if you'd like
        if not use_ox_states_in_mapping:
            structure.remove_oxidation_states()

        # ignore the original occupancies if you'd like (sometimes convenient)
        if not use_occ_in_mapping:
            els = self.els
            for el in els:
                structure = self.change_occ_for_el(el, {el: 1.0}, structure=structure)

        # figure out which elements have occupancy becoming 0
        disappearing_els = []
        for el_to_replace in species_mapping:
            if (len(species_mapping[el_to_replace]) == 1) and (
                list(species_mapping[el_to_replace].values())[0] == 0
            ):
                structure.remove_species(species=[el_to_replace])
                disappearing_els.append(el_to_replace)

        # remove these no longer existing elements
        if disappearing_els:
            for el in disappearing_els:
                del species_mapping[el]

        # replace species according to mapping
        if species_mapping:
            structure.replace_species(species_mapping)

        if structure.is_ordered:
            # if the replacement leads to an ordered structure, return it (in a dict)
            return {0: structure.as_dict()}
        else:
            # otherwise, need to order this partially occupied structure
            structools = StrucTools(structure, self.ox_states)
            return structools.get_ordered_structures(n_strucs=n_strucs)

    @property
    def spacegroup_info(self):
        """
        Returns:
            dict of spacegroup info with 'tight' or 'loose' symmetry tolerance
                tight means symprec = 0.01
                loose means symprec = 0.1
            e.g.,
                data['tight']['number'] returns spacegroup number with tight tolerance
                data['loose']['symbol'] returns spacegroup symbol with loose tolerance

        """
        data = {
            "tight": {"symprec": 0.01, "number": None, "symbol": None},
            "loose": {"symprec": 0.1, "number": None, "symbol": None},
        }
        for symprec in [0.01, 0.1]:
            sga = SpacegroupAnalyzer(self.structure, symprec=symprec)
            number = sga.get_space_group_number()
            symbol = sga.get_space_group_symbol()

            if symprec == 0.01:
                key = "tight"
            elif symprec == 0.1:
                key = "loose"

            data[key]["number"] = number
            data[key]["symbol"] = symbol

        return data

    def sg(self, number_or_symbol="symbol", loose_or_tight="loose"):
        """

        returns spacegroup number of symbol with loose or tight tolerance

        Args:
            number_or_symbol (str):
                whether to return the number or the symbol

            loose_or_tight (str):
                whether to use the loose or tight tolerance

        Returns:
            spacegroup number (int) or symbol (str) with loose or tight tolerance
        """
        sg_info = self.spacegroup_info
        return sg_info[loose_or_tight][number_or_symbol]

    def scale_structure(self, scale_factor, structure=None):
        """

        Isotropically scale a structure

        Args:
            scale_factor (float): fractional scaling of the structure volume
                - e.g., 1.2 will increase each lattice vector by 20%
                - e.g., 0.8 will make eaech lattice vector 80% of the initial length
                - e.g., 1.0 will do nothing

            structure (Structure, optional): structure to scale. Defaults to None.
                - if None will use self.structure fed to initialization

        Returns:
            pymatgen structure object
        """
        if not structure:
            structure = self.structure.copy()
        else:
            structure = structure.copy()
        orig_vol = structure.volume
        new_vol = orig_vol * scale_factor

        # scaling occurs only on the lattice (i.e., the top of POSCAR)
        structure.scale_lattice(new_vol)

        return structure


class SiteTools(object):
    """
    Purpose: make it a little easier to get site info from structures

    """

    def __init__(self, structure, index):
        """
        Args:
            structure (Structure)
                pymatgen Structure

            index (int)
                index of site in structure

        Returns:
            pymatgen Site object
        """
        structure = StrucTools(structure).structure
        self.site = structure[index]

    @property
    def site_dict(self):
        """
        Returns:
            dict of site info (from Pymatgen)
                {'species' : [{'element' : element, 'occu' : occupation, ...}, {...}}],
                 'abc' : fractional coordinates ([a, b, c])
                 'lattice' : Lattice object,
                 'properties' : dict (e.g., {'magmom' : 3})}
        """
        return self.site.as_dict()

    @property
    def coords(self):
        """
        Returns:
            array of fractional coordinates for site ([x, y, z])
        """
        return self.site.frac_coords

    @property
    def magmom(self):
        """
        Returns:
            magnetic moment for site (float) or None
        """
        props = self.site.properties
        if props:
            if "magmom" in props:
                return props["magmom"]
        return None

    @property
    def is_fully_occ(self):
        """
        Returns:
            True if site is fully occupied else False
        """
        return self.site.is_ordered

    @property
    def site_string(self):
        """
        unique string to represent a complex site

        Returns:
            occupation_element_oxstate__occupation_element_oxstate__... for each ion occupying a site
        """
        d = self.site_dict
        species = d["species"]
        ions = []
        for entry in species:
            el = entry["element"]
            if "oxidation_state" in entry:
                ox = float(entry["oxidation_state"])
            else:
                ox = None
            occ = float(entry["occu"])
            if ox:
                if ox < 0:
                    ox = str(abs(ox)) + "-"
                else:
                    ox = str(ox) + "+"
            else:
                ox = "0.0+"
            occ = str(occ)
            name_to_join = []
            for thing in [occ, el, ox]:
                if thing:
                    name_to_join.append(thing)
            name = "_".join(name_to_join)
            ions.append(name)
        if len(ions) == 1:
            return ions[0]
        else:
            return "__".join(ions)

    @property
    def ion(self):
        """
        Returns:
            the ion (element + oxidation state) occupying the site (str)
                None if > 1 ion
        """
        site_string = self.site_string
        if "__" in site_string:
            print("Multiple ions in site, returning None")
            return None

        return "".join([site_string.split("_")[1], site_string.split("_")[2]])

    @property
    def el(self):
        """
        Returns:
            just the element occupying the site (str)
                even if it has an oxidation state)

            None if more than one element occupies a site
        """
        site_string = self.site_string
        if "__" in site_string:
            print("Multiple ions in site, returning None")
            return None

        return site_string.split("_")[1]

    @property
    def ox_state(self):
        """
        Returns:
            oxidation state (float) of site
                averaged over all ions occupying the site
        """
        d = self.site_dict
        ox = 0
        species = d["species"]
        for entry in species:
            if entry["oxidation_state"]:
                ox += entry["oxidation_state"] * entry["occu"]
        return ox


def main():
    return


if __name__ == "__main__":
    main()



# from pydmclab.core.energies import ChemPots, FormationEnthalpy, FormationEnergy

# from .hpc import analyze, launch, submit, vasp
# from .utils import handy, plotting
# from .data import configs, features, reference_energies
# from .dev import energies, entries, grand, reactions



from pydmclab.core.energies import (
    ChemPots,
    FormationEnthalpy,
    FormationEnergy,
    DefectFormationEnergy,
)

from pydmclab.core.struc import StrucTools

from pydmclab.hpc.analyze import AnalyzeVASP
import os

""" 
Purpose:
    Thermodynamics runs the world :)
    
Basic idea:
    ChemPots: get elemental reference energies
    FormationEnthalpy: get formation enthalpy at 0 K
    FormationEnergy: get formation energy at T 
    DefectFormationEnergy: get defect formation energy
"""


def get_0K_chemical_potentials():
    """
    These are the elemental reference energies you need to reference your DFT calculated internal energies against to get 0 K formation enthalpies (energies)
        - different functionals (and standards) give different reference energies
        - critical to use compatible reference energies w/ your calculations
    """

    # assuming DMC standards for everything

    mus = ChemPots(functional="r2scan").chempots
    print("\n XC = metagga (T = 0 K)")
    print("mu_Al = %.2f eV/at" % mus["Al"])
    print("mu_N = %.2f eV/at" % mus["N"])

    mus = ChemPots(functional="pbe").chempots
    print("\n XC = gga (T = 0 K)")
    print("mu_Al = %.2f eV/at" % mus["Al"])
    print("mu_N = %.2f eV/at" % mus["N"])


def get_T_dependent_chemical_potentials():
    """
    we can also use experimental data to get chemical potentials at T
        - from Barin's thermochemical data of pure substances
    """
    print("\n T-dependent mus")
    allowed_Ts = list(range(300, 2100, 100))
    for T in allowed_Ts:
        # note functional and standard don't matter for T > 0
        print("mu_N(%i K) = %.3f eV/at" % (T, ChemPots(temperature=T).chempots["N"]))


def get_chemical_potentials_with_gas_activity():
    """
    we can also account for the effects of gas concentration (activity) in the atmosphere
        - according to dmu = RTln(p_gas/p_gas,ref)
    """
    print("\n pN2-dependent mus")
    pN2s = [10**i for i in range(-6, 1)][::-1]
    T = 1000
    for pN2 in pN2s:
        print(
            "mu_N(1000 K, pN2 = %.0e atm) = %.3f eV/at"
            % (
                pN2,
                ChemPots(temperature=1000, partial_pressures={"N": pN2}).chempots["N"],
            )
        )


def get_mus_with_custom_shift():
    """
    we can also just specify whatever we want or shift chemical potentials as needed
    """
    print("\n custom mu shifts")
    mus = ChemPots(
        functional="r2scan", standard="dmc", user_chempots={"N": 10000}
    ).chempots
    print("user set mu_N = %.2f eV/at" % mus["N"])

    mus = ChemPots(functional="r2scan", standard="dmc", user_dmus={"N": 0.1}).chempots
    print("user shifted mu_N up by 0.1 eV/at to %.2f eV/at" % mus["N"])


def get_formation_enthalpy_at_0K():
    """
    these are DFT formation energies (at 0 K)
        Bartel npj 2019 (and others) shows these agree well with dHf(298 K)
    """
    print("\n formation enthalpy (internal energy)")
    formula = "SiC"
    E_DFT = -7.5308  # per atom
    mus = ChemPots(functional="pbe").chempots

    Ef = FormationEnthalpy(formula, E_DFT, mus).Ef
    print("Ef = %.2f eV/atom" % Ef)


def get_T_dependent_formation_energy_with_vibrational_entropy():
    """
    Bartel et al Nature Comms 2019 allows us to get DFT energies compatbile w/ finite T chemical potentials
        by approximating vibrational entropy
    """
    print("\n formation energy (Gibbs free energy from vibrational entropy)")
    data_dir = "../data/test_data/vasp/AlN"

    formula = "AlN"

    Ef = -1.6  # eV/atom (you could also compute this with FormationEnthalpy)
    path_to_poscar = os.path.join(data_dir, "CONTCAR")
    structure = StrucTools(path_to_poscar).structure

    for T in [0, 300, 900, 1500]:
        mus = ChemPots(temperature=T).chempots
        dGf = FormationEnergy(formula, Ef, mus, structure, include_Svib=True).dGf(T)
        print("dGf(%i K) = %.2f eV/atom" % (T, dGf))


def get_T_dependent_formation_energy_with_configurational_entropy():
    """
    we can also account for configurational entropy using a simple ideal mixing approximation
        S_config = -k_B * (x * ln(x) + (1-x) * ln(1-x))
        - note: this neglects any short-range order and assumes fully random occupations
    """
    print("\n formation energy (Gibbs free energy from configurational entropy)")
    formula = "Ag3Au"
    Ef = -0.041  # eV/atom (you could also compute this with FormationEnthalpy)

    for T in [0, 300, 900, 1500]:
        mus = ChemPots(temperature=T).chempots
        dGf = FormationEnergy(
            formula,
            Ef,
            mus,
            include_Svib=False,
            include_Sconfig=True,
            x_config=0.75,  # 3 Ag and 1 Au can occupy the same site
            n_config=1,  # one site having mixed occupation
        ).dGf(T)
        print("dGf(%i K) = %.2f eV/atom" % (T, dGf))


def get_defect_formation_energy():
    """
    defect formation energies tell me the energy gain or penalty of creating defects
        - vacancies, interstitials, substitutions
        - antisites not implemented yet...
    """
    print("\n now a defect")
    pristine_formula = "GaN"
    defect_formula = "AlGa3N4"

    Ef_pristine = -0.668  # eV/atom
    Ef_defect = -0.896  # eV/atom

    Eg_pristine = 1.74  # eV

    q = 0  # neutral substitutional defect

    charge_correction = 0  # not charged

    shared_el = "N"  # for normalizing formula

    mus = ChemPots(functional="pbe", standard="mp").chempots  # need mus (for Ga and Al)

    E_Fermi = 0  # neutral defect, so let's evaluate at VBM

    dfe = DefectFormationEnergy(
        formula_defect=defect_formula,
        formula_pristine=pristine_formula,
        E_defect=Ef_defect,
        E_pristine=Ef_pristine,
        Eg_pristine=Eg_pristine,
        charge_defect=q,
        charge_correction=charge_correction,
        shared_el_for_basis=shared_el,
        chempots=mus,
    ).Efs[E_Fermi]

    print("defect formation energy for Al_Ga in GaN = %.2f eV/atom" % dfe)


def main():
    get_0K_chemical_potentials()
    get_T_dependent_chemical_potentials()
    get_chemical_potentials_with_gas_activity()
    get_mus_with_custom_shift()
    get_formation_enthalpy_at_0K()
    get_T_dependent_formation_energy_with_vibrational_entropy()
    get_T_dependent_formation_energy_with_configurational_entropy()
    get_defect_formation_energy()
    return


if __name__ == "__main__":
    main()



from pydmclab.core.struc import StrucTools, SiteTools
import itertools
import random
from pymatgen.core.structure import Structure
from pymatgen.transformations.site_transformations import (
    ReplaceSiteSpeciesTransformation,
)
from pymatgen.analysis.structure_matcher import StructureMatcher


class MagTools(object):
    """
    For generating magnetic orderings for structures
    """

    def __init__(
        self,
        structure,
        max_afm_combos=100,
        afm_spins=(-5, 5),
        fm_spins=(0.6, 5),
        randomize_afm=True,
        treat_as_nm=[],
    ):
        """
        Args:
            structure (Structure):
                pymatgen Structure object

            max_afm_combos (int):
                maximum number of AFM spin configurations to generate
                    this is useful when there are a tremendous number of plausible AFM orderings (e.g., for structures w/ many magnetic elements)
                    100 is usually reasonable

            afm_spins (tuple):
                low and high spin for AFM initialization
                    -5 and 5 are usually safe initializations

            fm_spins (tuple):
                zero and non-zero spin for FM initialization
                    0.6 and 5 are usually safe initializations

            randomize_afm (bool):
                randomize AFM spin configurations
                    randomization occurs in two different steps
                        1) when AFM combinations are enumerated,
                            if the number of combinations is greater than max_afm_combos,
                            then we randomly select max_afm_combos combinations
                        2) when the unique AFM structures are generated,
                            we randomly order them to avoid having the same "kinds" of AFM orderings always appearing first
                    random seeds are being used so both operations should be deterministic

            treat_as_nm (list):
                list of elements to treat as non-magnetic
                    e.g., if you want to only explore various initial configurations for other element(s)
        """
        if isinstance(structure, dict):
            structure = Structure.from_dict(structure)

        # purge oxidation states (makes life easier later)
        structure.remove_oxidation_states()
        self.structure = structure
        self.max_afm_combos = max_afm_combos
        self.fm_spins = fm_spins
        self.afm_spins = afm_spins
        self.randomize_afm = randomize_afm
        self.treat_as_nm = treat_as_nm

    @property
    def magnetic_ions(self):
        """
        Aggregated from MP + matminer (**propably a better list to use somewhere**)

        from_MP = https://github.com/materialsproject/pymatgen/blob/master/pymatgen/analysis/magnetism/default_magmoms.yaml
        from_matminer = https://github.com/hackingmaterials/matminer/blob/main/matminer/featurizers/composition/element.py

        Returns:
            list of elements (str) that could be magnetic
                note removes those you've said to "treat_as_nm"
        """
        from_matminer = [
            "Ti",
            "V",
            "Cr",
            "Mn",
            "Fe",
            "Co",
            "Ni",
            "Cu",
            "Nb",
            "Mo",
            "Tc",
            "Ru",
            "Rh",
            "Pd",
            "Ag",
            "Ta",
            "W",
            "Re",
            "Os",
            "Ir",
            "Pt",
        ]
        from_MP = [
            "Co",
            "Cr",
            "Fe",
            "Mn",
            "Mo",
            "Ni",
            "V",
            "W",
            "Ce",
            "Eu",
            "Ti",
            "V",
            "Cr",
            "Mn",
            "Fe",
            "Co",
            "Ni",
            "Cu",
            "Pr",
            "Nd",
            "Pm",
            "Sm",
            "Gd",
            "Tb",
            "Dy",
            "Ho",
            "er",
            "Tm",
            "Yb",
            "Np",
            "Ru",
            "Os",
            "Ir",
            "U",
        ]
        magnetic = list(set(from_matminer + from_MP))
        treat_as_nm = self.treat_as_nm
        return sorted([el for el in magnetic if el not in treat_as_nm])

    @property
    def magnetic_ions_in_struc(self):
        """
        list of elements (str) in structure that are magnetic
        """
        els = StrucTools(self.structure).els
        magnetic_ions = self.magnetic_ions
        return sorted(list(set([el for el in els if el in magnetic_ions])))

    @property
    def could_be_magnetic(self):
        """
        True if any magnetic ions are in structure else False
        """
        return True if len(self.magnetic_ions_in_struc) > 0 else False

    @property
    def could_be_afm(self):
        """
        True if there are at least two magnetic sites in structure and the number of magnetic sites is even
        """
        if not self.could_be_magnetic:
            return False

        magnetic_ions = self.magnetic_ions_in_struc
        magnetic_sites = 0
        structure = self.structure
        for idx in range(len(structure)):
            el = SiteTools(structure, idx).el
            if el in magnetic_ions:
                magnetic_sites += 1
        if magnetic_sites > 1:
            if magnetic_sites % 2 == 0:
                return True

    @property
    def get_nonmagnetic_structure(self):
        """
        Returns nonmagnetic Structure with magmoms of zeros
        """
        s = self.structure
        magmom = [0 for i in range(len(s))]
        s_tmp = s.copy()
        s_tmp.add_site_property("magmom", magmom)
        return s_tmp

    @property
    def get_ferromagnetic_structure(self):
        """
        Returns Structure with all magnetic ions ordered ferromagnetically
            - nonmagnetic ions are given spin = spins[0] (default: 0.6)
            - magnetic ions are given spin = spins[1] (default: 5)
        """
        spins = self.fm_spins
        magnetic_ions_in_struc = self.magnetic_ions_in_struc
        if len(magnetic_ions_in_struc) == 0:
            return None
        s = self.structure
        magmom = [
            spins[0] if s[i].species_string not in magnetic_ions_in_struc else spins[1]
            for i in range(len(s))
        ]
        s_tmp = s.copy()
        s_tmp.add_site_property("magmom", magmom)
        return s_tmp

    @property
    def get_antiferromagnetic_structures(self):
        """
        This is a chaotic way to get antiferromagnetic configurations
            - but it doesn't require enumlib interaction with pymatgen
            - it seems reasonably efficient, might break down for large/complex structures
            - note 1: it has no idea which configurations are "most likely" to be low energy
            - note 2: it may require execution on compute nodes

        Basic workflow:
            - start from the NM structure
            - for all sites containing ions in magnetic_ions
                - generate all possible combinations of 0 (spin down) or 1 (spin up) for each site
                    - if I had four sites w/ mag ions this might be: [(0,0,0,1), (0,0,1,1), ...]
                - retain only the combinations that have average = 0.5 (ie half spin down, half spin up)
            - now apply all these combinations to the structure
                - generate a new structure for each combination that puts max(spin) on sites with 1 and min(spin) on sites with 0
            - now figure out which newly generated structures are symmetrically distinct
                - change the identities of sites that are spin up/down using oxidation state surrogate
                    - these ox states aren't physically meaningful, just a placeholder
                    - spin up: 8+, spin down: 8-
            - now use StructureMatcher to find unique structures to return

        Returns:
            list of unique Structure objects with antiferromagnetic ordering
                - exhaustive if len(combos) <= max_combos
                - no idea which are most likely to be low energy
                - reasonable to randomly sample if a very large list

            in general, probably unnecessary to calculate more than ~2-10 of these per structure
        """

        # parameters that could be args...
        spins = self.afm_spins  # magnitudes of high/low spin
        max_combos = (
            self.max_afm_combos
        )  # max number of combinations to try (for cases with very many possible combos)

        # which ions in structure are magnetic:
        magnetic_ions_in_struc = self.magnetic_ions_in_struc
        if len(magnetic_ions_in_struc) == 0:
            return None

        strucs_with_magmoms = []
        s = self.get_nonmagnetic_structure

        # get sites w/ magnetic ions
        magnetic_sites = [
            i for i in range(len(s)) if s[i].species_string in magnetic_ions_in_struc
        ]

        print("%i magnetic sites" % len(magnetic_sites))
        # enumerate all possible ways to yield afm ordering
        combos = itertools.product(range(len(spins)), repeat=len(magnetic_sites))

        combos = [c for c in combos if sum(c) == 0.5 * len(magnetic_sites)]
        print("%i afm combos" % len(combos))

        # randomly reduce list if too big for practical usage
        if len(combos) > max_combos:
            combos = random.Random(0).sample(combos, max_combos)
            print("reduced to: %i afm combos" % len(combos))

        # decorate structures w/ magmoms for all afm orderings
        for j in range(len(combos)):
            c = combos[j]
            magnetic_moments = [spins[c[i]] for i in range(len(c))]
            site_idxs_to_magmom = dict(zip(magnetic_sites, magnetic_moments))
            for i in range(len(s)):
                if i not in magnetic_sites:
                    site_idxs_to_magmom[i] = 0
            magmom = [site_idxs_to_magmom[i] for i in range(len(s))]
            s_tmp = s.copy()
            s_tmp.add_site_property("magmom", magmom)
            # print(s_tmp.site_properties)
            strucs_with_magmoms.append(s_tmp)
        print("made strucs")

        # replace spin-up and spin-down sites with new species for symmetry matching
        fake_strucs = []
        for struc in strucs_with_magmoms:
            magmom = struc.site_properties["magmom"]
            spin_up = [i for i in magnetic_sites if magmom[i] == spins[1]]
            spin_down = [i for i in magnetic_sites if magmom[i] == spins[0]]
            indices_species_map = {
                idx: struc[idx].species_string + "8+"
                if idx in spin_up
                else struc[idx].species_string + "8-"
                for idx in spin_up + spin_down
            }
            rsst = ReplaceSiteSpeciesTransformation(indices_species_map)
            s_tmp = rsst.apply_transformation(struc)
            fake_strucs.append(s_tmp)
        fake_strucs_dict = dict(zip(list(range(len(fake_strucs))), fake_strucs))

        matcher = StructureMatcher()
        groups = matcher.group_structures(
            [fake_strucs_dict[i] for i in fake_strucs_dict]
        )
        unique_strucs = []
        for g in groups:
            s = g[0]
            s.remove_oxidation_states()
            unique_strucs.append(s)

        #        print('%i unique afm structures' % len(unique_strucs))

        if self.randomize_afm:
            random.Random(1).shuffle(unique_strucs)

        print("%i unique afm structures" % len(unique_strucs))
        return unique_strucs

    @property
    def get_afm_magmoms(self):
        """
        Returns:
            dict of magmoms for each AFM ordering for a given structure
                {idx of configuration (int) : magmoms (list)}
        """
        afm_strucs = self.get_antiferromagnetic_structures
        magmoms = {}
        if afm_strucs:
            for i in range(len(afm_strucs)):
                magmoms[i] = afm_strucs[i].site_properties["magmom"]

        return magmoms


def main():
    return


if __name__ == "__main__":
    main()



from pydmclab.core.query import MPQuery
from pydmclab.core.struc import StrucTools
from pydmclab.utils.handy import read_json, write_json
import os

""" 
Purpose:
    get data from the Materials Project database
    
Note:
    uses "old" API
"""

API_KEY = "N3KdATtMmcsUL94g"
DATA_DIR = os.path.join("output", "query")
CHEMSYS = "Li-Mn-F-S"


def demo_get_groundstate_hull_data_for_chemsys(
    chemsys=CHEMSYS,
    remake=False,
    data_dir=DATA_DIR,
    api_key=API_KEY,
):
    """
    Args:
        chemsys (str)
            chemical system to query ('-'.join([elements]))

        only_gs (bool)
            if True, remove non-ground state polymorphs from MP query
                good practice to do this before doing hull analysis b/c non-gs polymorphs trivially have Ehull=Ehull_gs + dE_polymorph-gs

        dict_key (str)
            key to use to orient dictionary of MPQuery results
                'cmpd' is the default behavior in MPQuery, meaning we get a dictionary that looks like {CHEMICAL_FORMULA : {DATA}}

        remake (bool)
            if True, re-query MP

    Returns:
        gs (dict): dictionary of ground state data from MPQuery for chemsys
    """
    fjson = os.path.join(data_dir, "query_gs_all_" + chemsys + ".json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    mpq = MPQuery(api_key)
    out = mpq.get_data_for_comp(comp=chemsys)

    return write_json(out, fjson)


def reorient_gs_data_from_ID_keys_to_cmpd_keys(d_gs):
    """
    Args:
        d_gs (dict): dictionary of ground state data from MPQuery for chemsys
            {MP_ID : {DATA}}

    Returns:
        d_gs_cmpd (dict):
            reoriented so that keys are chemical formula instead of MP ID
    """
    d_gs_cmpd = {}
    for k in d_gs:
        d_gs_cmpd[d_gs[k]["cmpd"]] = d_gs[k]
    return d_gs_cmpd


def demo_entry_retrieval(mpid="mp-530748"):
    """
    Grab a calculated entry from MP
    """
    mpq = MPQuery(API_KEY)
    properties = list(mpq.typical_properties) + ["band_gap"]
    entry = mpq.get_entry_by_material_id(
        material_id=mpid,
        properties=properties,
        incl_structure=True,
        conventional=False,
        compatible_only=True,
    )
    print(entry)
    return entry


def demo_structure_retrieval(mpid="mp-1094961"):
    mpq = MPQuery(API_KEY)
    s = mpq.get_structure_by_material_id(material_id=mpid)

    print("\nThe formula for this structure is %s" % StrucTools(s).formula)
    print("The first site in the structure is %s" % s[0])

    return s


def demo_vaspinput_retrieval(mpid="mp-1938"):
    inputs = MPQuery(API_KEY).get_vasp_inputs(material_id=mpid)
    print("\n The %s for this calculation was %s" % ("ALGO", inputs["incar"]["ALGO"]))
    return inputs


def main():
    d_gs = demo_get_groundstate_hull_data_for_chemsys(remake=True)
    d_gs_cmpd = reorient_gs_data_from_ID_keys_to_cmpd_keys(d_gs)
    entry = demo_entry_retrieval()
    structure = demo_structure_retrieval()
    vasp_inputs = demo_vaspinput_retrieval()
    d_gs, d_gs_cmpd = None, None
    return d_gs, d_gs_cmpd, entry, structure, vasp_inputs


if __name__ == "__main__":
    out = main()



from pymatgen.core.composition import Composition
import numpy as np


class CompTools(object):
    def __init__(self, formula):
        """
        Args:
            formula (str) - chemical formula

        Returns:
            formula
        """
        self.formula = formula

    @property
    def clean(self):
        """
        Returns:
            formula (str) that has been:
                - sorted by elements
                - parentheses removed
                - fractions --> integers

        In general, we want to use this for maximum consistency
        """
        # reduce
        formula = Composition(self.formula).reduced_formula

        # alphabetize
        formula = Composition(formula).alphabetical_formula

        # expand if it has decimals
        if "." in formula:
            formula = Composition(formula).get_integer_formula_and_factor()[0]

        # alphabetize
        formula = Composition(formula).alphabetical_formula

        # remove parentheses
        formula = Composition(formula).to_pretty_string()

        # remove "2" for diatomic elements
        if (len(formula) <= 3) and (formula[-1] == "2"):
            formula = formula.replace("2", "1")
        return formula

    @property
    def pretty(self):
        """
        Returns:
            formula (str) that is visually pleasing (chemically meaningful)
                - note: reduces the formula
        """
        return Composition(self.formula).reduced_formula

    @property
    def amts(self):
        """
        Returns:
            dictionary of elements (str) and their amounts (float)
                - note: starts with "clean" formula
                {el (str) : amt (float)}

            CompTools('Al4O6').amts = {'Al' : 2, 'O' : 3}
        """

        return Composition(self.clean).get_el_amt_dict()

    def mol_frac(self, el):
        """
        Returns:
            the molar fraction (float) of an element (str)
                - note: starts with "clean" formula

            CompTools('Al4O6').mol_frac('Al') = 0.4
        """
        return Composition(self.clean).get_atomic_fraction(el)

    def stoich(self, el):
        """
        Returns:
            the stoichiometry of an element (int)
                - note: starts with "clean" formula
            e.g., if CompTools(c).clean == 'Al2Mg1O4', then CompTools(c).stoich('O') = 4
        """
        stoich = self.mol_frac(el) * self.n_atoms
        stoich = np.round(stoich, 0)
        return int(stoich)

    @property
    def chemsys(self):
        """
        Returns:
            chemical system (str) of the formula
                - sorted
                - elements (str) joined by "-"

            CompTools('Al4O6').chemsys = 'Al-O'
        """
        return Composition(self.clean).chemical_system

    @property
    def els(self):
        """
        Returns:
            list of elements (str) in the formula
                - sorted

            CompTools('Al4O6').els = ['Al', 'O']
        """
        return list(sorted(self.chemsys.split("-")))

    @property
    def n_els(self):
        """
        Returns:
            number of elements (int) in the formula

            CompTools('Al4O6').n_els = 2
        """
        return len(self.els)

    @property
    def n_atoms(self):
        """
        Returns:
            number of atoms (int) in the formula
                - note: starts with "clean" formula

            CompTools('Al4O6').n_atoms = 5
        """
        return np.sum(list(self.amts.values()))

    def label_for_plot(self, el_order=None):
        """
        @NOTE:
            this is redundant w/ pydmclab.plotting.utils.get_label ?

        @FEATURE:
            - would be great if this worked for decimal formulas

        Args:
            el_order (list) - order of elements (str) for plotting

        Returns:
            label (str) for plotting (includes $ for subscripts)
        """
        # formula = self.clean if reduce else self.formula
        if not el_order:
            el_order = self.els
        amts = self.amts
        label = r"$"
        for el in el_order:
            if el in amts:
                n_el = amts[el]
                if n_el == 1:
                    label += el
                elif n_el > 1:
                    if int(n_el) - n_el == 0:
                        label += el + "_{%s}" % str(int(n_el))
                    else:
                        label += el + "_{%.1f}" % float(n_el)
        label += "$"
        return label


def main():
    return


if __name__ == "__main__":
    main()



import numpy as np
import math
from itertools import combinations

from pydmclab.data.thermochem import (
    mp2020_compatibility_dmus,
    mus_at_0K,
    mus_at_T,
    mus_from_mp_no_corrections,
    mus_from_bartel2019_npj,
)
from pydmclab.data.features import atomic_masses
from pydmclab.core.comp import CompTools
from pydmclab.core.struc import StrucTools

from pymatgen.analysis.reaction_calculator import Reaction
from pymatgen.core.composition import Composition


class ChemPots(object):
    """
    return dictionary of chemical potentials {el : chemical potential (eV/at)} based on user inputs
    """

    def __init__(
        self,
        temperature=0,
        functional="pbe",
        standard="dmc",
        partial_pressures={},  # atm
        diatomics=["H", "N", "O", "F", "Cl"],
        oxide_type="oxide",
        user_chempots={},
        user_dmus={},
    ):
        """
        Args:
            temperature (int)
                temperature in Kelvin
                    if T > 0, will use experimental data from Barin's thermochemical data of pure substances

            functional (str)
                explicit functional for DFT claculations (don't include + in name)
                    currently supports "pbe" (for GGA), "pbeu" (for GGA+U), "r2scan" (for meta-GGA)

            standard (str)
                standard for DFT calculations
                    currently supports "dmc" (for DMC) and "mp" (for Materials Project)

            partial_pressures (dict)
                {el (str) : partial pressure (atm)}
                    adjusts chemical potential of gaseous species based on RTln(p/p0)
                        where p0 = 1 atm

            diatomics (list)
                list of diatomic elements
                    if el is in diatomics, will use 0.5 * partial pressure effect on mu

            oxide_type (str)
                type of oxide
                    this only affects MP formation energies
                    they use different corrections for oxides, peroxides, and superoxides

            user_chempots (dict)
                {el (str) : chemical potential (eV/at)}
                    specifies the chemical potential you want to use for el
                    will override everything

            user_dmus (dict)
                {el (str) : delta_mu (eV/at)}
                    specifies the change in chemical potential you want to use for el
                    will be added on top of everything except user_chempots
        """
        self.temperature = temperature
        self.functional = functional
        self.standard = standard
        self.partial_pressures = partial_pressures
        self.diatomics = diatomics
        self.oxide_type = oxide_type
        self.user_dmus = user_dmus
        self.user_chempots = user_chempots

    @property
    def apply_mp_corrections(self):
        """
        Returns:
            updates user_dmus to include MP corrections
        """
        user_dmus = self.user_dmus.copy()

        # load the data extracted in ~2022 (from 2020 compatibility scheme)
        mp_dmus = mp2020_compatibility_dmus()

        # shift the anions MP wants to shift
        for el in mp_dmus["anions"]:
            user_dmus[el] = -mp_dmus["anions"][el]

        # apply U corrections
        if self.functional == "pbeu":
            for el in mp_dmus["U"]:
                user_dmus[el] = -mp_dmus["U"][el]

        # apply different kinds of oxide corrections
        if self.oxide_type == "peroxide":
            user_dmus[el] = -mp_dmus["peroxide"]["O"]
        elif self.oxide_type == "superoxide":
            user_dmus[el] = -mp_dmus["superoxide"]["O"]

        self.user_dmus = user_dmus.copy()

    @property
    def chempots(self):
        """
        Returns:
            dictionary of chemical potentials
                {el : chemical potential (eV/at)} based on user inputs
        """
        T = self.temperature
        standard, functional = self.standard, self.functional
        if T == 0:
            # use DFT data at 0 K
            if (standard == "dmc") or (functional in ["scan", "r2scan"]):
                # use DMC data
                all_mus = mus_at_0K()
                els = sorted(list(all_mus[functional].keys()))
                mus = {el: all_mus[functional][el]["mu"] for el in els}
            else:
                # use MP data
                mus = mus_from_mp_no_corrections()
        else:
            # use experimental data at T > 0 K
            allowed_Ts = list(range(300, 2100, 100))
            if T not in allowed_Ts:
                raise ValueError("Temperature must be one of %s" % allowed_Ts)
            all_mus = mus_at_T()
            mus = all_mus[str(T)].copy()

        # apply partial pressure correction for activity of gaseous elements
        partial_pressures = self.partial_pressures
        diatomics = self.diatomics
        R = 8.6173303e-5  # eV/K

        if partial_pressures:
            for el in partial_pressures:
                if el in diatomics:
                    # correct diatomics b/c they exist as O2 yet their mu is stored as O
                    factor = 1 / 2
                else:
                    factor = 1
                mus[el] += R * T * factor * np.log(partial_pressures[el])

        if (standard == "mp") and (T == 0):
            # apply MP corrections if needed
            self.apply_mp_corrections

        user_dmus = self.user_dmus
        if user_dmus:
            # apply any dmus if needed
            for el in user_dmus:
                mus[el] += user_dmus[el]

        user_chempots = self.user_chempots
        if user_chempots:
            # specify any mus directly
            for el in user_chempots:
                mus[el] = user_chempots[el]

        return mus.copy()


class FormationEnthalpy(object):
    """
    For computing formation energies (~equivalently enthalpies) at 0 K
    """

    def __init__(
        self,
        formula,
        E_DFT,
        chempots,
    ):
        """
        Args:
            formula (str)
                chemical formula

            E_DFT (float)
                DFT energy (eV/atom)
                    per atom should mean per CompTools(formula).n_atoms

            chempots (dict)
                {el (str) : chemical potential (eV/at)}
                    probably generated using ChemPots(...).chempots

        """
        self.formula = CompTools(formula).clean
        self.E_DFT = E_DFT
        self.chempots = chempots

    @property
    def weighted_elemental_energies(self):
        """
        Returns:
            weighted elemental energies (eV per formula unit)

                for Al2O3, this would be 2 * mu_Al + 3 * mu_O
        """
        mus = self.chempots
        els_to_amts = CompTools(self.formula).amts
        for el in els_to_amts:
            if (el not in mus) or not mus[el]:
                raise ValueError('No chemical potential for "%s"' % el)
        return np.sum([mus[el] * els_to_amts[el] for el in els_to_amts])

    @property
    def Ef(self):
        """
        Returns:
            formation energy at 0 K (eV/atom)
                the formation reaction for AxBy is xA + yB --> AxBy
                this reaction energy is computed on a per formula unit (molar) basis
                then divided by the number of atoms (x + y) to get eV/atom formation energies
        """
        formula = self.formula
        n_atoms = CompTools(formula).n_atoms
        weighted_elemental_energies = self.weighted_elemental_energies
        E_per_fu = self.E_DFT * n_atoms
        return (1 / n_atoms) * (E_per_fu - weighted_elemental_energies)


class FormationEnergy(object):
    """
    This class is for computing formation energies at T > 0 K

    By default, uses the Bartel2018 model for vibrational entropy: https://doi.org/10.1038/s41467-018-06682-4

    """

    def __init__(
        self,
        formula,
        Ef,
        chempots,
        structure=False,
        atomic_volume=False,
        x_config=None,
        n_config=1,
        include_Svib=True,
        include_Sconfig=False,
    ):
        """
        Args:
            formula (str)
                chemical formula

            Ef (float)
                DFT formation enthalpy at 0 K (eV/at)
                    or any formation enthalpy at T <= 298 K

            chempots (dict)
                {el (str) : chemical potential (eV/at)}
                    probably generated using ChemPots.chempots

            structure (Structure)
                pymatgen structure object (or Structure.as_dict() or path to structure file)
                    either structure or atomic_volume needed for vibrational entropy calculation

            atomic_volume (float)
                atomic volume (A^3/atom)
                    either structure or atomic_volume needed for vibrational entropy calculation

            x_config (float)
                partial occupancy parameter to compute configurational entropy
                    needed to compute configurational entropy [x in xlnx + (1-x)ln(1-x))]

            n_config (int)
                number of inequivalent sites exhibiting ideal solution behavior
                    this would be one if I have one site that is partially occupied by two ions
                    this would be two if I have two sites that are each partially occupied by two ions

            include_Svib (bool)
                whether to include vibrational entropy (Bartel model)

            include_Sconfig (bool)
                whether to include configurational entropy (ideal mixing model)
        """
        self.formula = CompTools(formula).clean
        self.Ef = Ef
        self.chempots = chempots
        if structure:
            structure = StrucTools(structure).structure
        self.structure = structure
        self.atomic_volume = atomic_volume
        self.include_Svib = include_Svib
        self.include_Sconfig = include_Sconfig
        self.x_config = x_config
        self.n_config = n_config

        if include_Svib:
            if not structure and not atomic_volume:
                raise ValueError(
                    "Must provide structure or atomic volume to compute Svib"
                )

        if include_Sconfig:
            if not (x_config and n_config):
                if x_config != 0:
                    raise ValueError(
                        "Must provide x_config and n_config to compute Sconfig"
                    )

    @property
    def weighted_elemental_energies(self):
        """
        Returns:
            weighted elemental energies (eV per formula unit)
        """
        mus = self.chempots
        els_to_amts = CompTools(self.formula).amts
        for el in els_to_amts:
            if (el not in mus) or not mus[el]:
                raise ValueError('No chemical potential for "%s"' % el)
        return np.sum([mus[el] * els_to_amts[el] for el in els_to_amts])

    @property
    def reduced_mass(self):
        """
        Returns weighted reduced mass of composition
            needed if include_Svib = True (Chris B Nature Comms 2019)
        """
        names = CompTools(self.formula).els
        els_to_amts = CompTools(self.formula).amts
        nums = [els_to_amts[el] for el in names]
        mass_d = atomic_masses()
        num_els = len(names)
        num_atoms = np.sum(nums)
        denom = (num_els - 1) * num_atoms
        if denom <= 0:
            print("descriptor should not be applied to unary compounds (elements)")
            return np.nan
        masses = [mass_d[el] for el in names]
        good_masses = [m for m in masses if not math.isnan(m)]
        if len(good_masses) != len(masses):
            for el in names:
                if math.isnan(mass_d[el]):
                    print("I dont have a mass for %s..." % el)
                    return np.nan
        else:
            pairs = list(combinations(names, 2))
            pair_red_lst = []
            for i in range(len(pairs)):
                first_elem = names.index(pairs[i][0])
                second_elem = names.index(pairs[i][1])
                pair_coeff = nums[first_elem] + nums[second_elem]
                pair_prod = masses[first_elem] * masses[second_elem]
                pair_sum = masses[first_elem] + masses[second_elem]
                pair_red = pair_coeff * pair_prod / pair_sum
                pair_red_lst.append(pair_red)
            return np.sum(pair_red_lst) / denom

    @property
    def S_config(self):
        """
        configurational entropy from ideal mixing model (float, eV/atom/K)
            no short range order
            completely random occupation



        -kB * n_config * (x_config * ln(x_config) + (1-x_config) * ln(1-x_config))
        """
        x, n = self.x_config, self.n_config
        if x in [0, 1]:
            return 0
        kB = 8.617e-5  # eV/K
        S_config = (-kB * n * (x * np.log(x) + (1 - x) * np.log(1 - x))) / CompTools(
            self.formula
        ).n_atoms
        return S_config

    def dGf(self, temperature):
        """
        Args:
            temperature (int)
                temperature (K)

        Returns:
            formation energy at temperature (eV/at)
                see Chris B Nature Comms 2019
        """
        T = temperature
        Ef_0K = self.Ef
        if T == 0:
            # use 0 K formation energy
            return Ef_0K
        if self.include_Svib:
            m = self.reduced_mass
            if self.atomic_volume:
                V = self.atomic_volume
            elif self.structure:
                V = self.structure.volume / len(self.structure)
            else:
                raise ValueError("Need atomic volume or structure to compute G(T)")

            Gd_sisso = (
                (-2.48e-4 * np.log(V) - 8.94e-5 * m / V) * T + 0.181 * np.log(T) - 0.882
            )
            weighted_elemental_energies = self.weighted_elemental_energies
            G = Ef_0K + Gd_sisso
            n_atoms = CompTools(self.formula).n_atoms

            dGf = (1 / n_atoms) * (G * n_atoms - weighted_elemental_energies)

        if self.include_Sconfig:
            if not self.include_Svib:
                # start from 0 K formation energy
                dGf = Ef_0K
            dGf += -T * self.S_config

        return dGf


class DefectFormationEnergy(object):
    def __init__(
        self,
        E_pristine,
        formula_pristine,
        Eg_pristine,
        E_defect,
        formula_defect,
        charge_defect,
        shared_el_for_basis,
        chempots,
        charge_correction,
        gap_discretization=0.1,
    ):
        """
        Args:
            E_pristine (float)
                DFT total energy (or formation energy) of pristine compound (eV/atom)

            formula_pristine (str)
                formula of pristine compound

            Eg_pristine (float)
                band gap of pristine compound (eV)

            E_defect (float)
                DFT total energy (or formation energy) of defect-containing compound (eV/atom)

            formula_defect (str)
                formula of defect-containing compound

            charge_defect (int)
                charge of defect-containing compound

            shared_el_for_basis (str)
                element shared between pristine and defect-containing compounds
                    - for normalizing formula units
                    - e.g., if pristine = Al2O3 and defect = Al4O5S1
                        - shared_el_for_basis = "Al"
                        - so that I know to use Al4O6 for pristine for calculations..

            chempots (dict)
                chemical potentials (eV/atom)
                    {el (str) :
                        chempot (float)
                    for any el exchanged between defect and pristine}

            charge_correction (float)
                charge correction (eV/cell)
                    this gets added before dividing by number of defects

            gap_discretization (float)
                how fine of a grid do you want to calculate the formation energy over
                smaller numbers = more grid points

        Returns:
            shared_el_for_basis
            chempots
            charge_correction
            gap_discretization
            pristine (dict)
                {'E' : total energy per atom in clean formula (eV/atom),
                 'formula' : the clean formula,
                 'Eg' : band gap (eV),
                 'multiplier' : how many times the clean formula is multiplied to get the defect-containing formula.
                 'basis_formula' : the formula of the basis (e.g., Al4O6 for Al2O3 if the defect if Al4O5S1)}
            defect (dict)
                {'E' : total energy per atom in defect-containing formula (eV/atom),
                 'formula' : the clean defect-containing formula,
                 'q' : charge of defect-containing formula}

        """
        # clean formulas
        formula_pristine = CompTools(formula_pristine).clean
        formula_defect = CompTools(formula_defect).clean

        # make dicts for pristine and defect
        pristine = {"E": E_pristine, "formula": formula_pristine, "Eg": Eg_pristine}
        defect = {"E": E_defect, "formula": formula_defect, "q": charge_defect}

        self.shared_el_for_basis = shared_el_for_basis
        self.chempots = chempots
        self.charge_correction = charge_correction

        # map pristine to defect basis
        pristine_multiplier = CompTools(formula_defect).stoich(
            shared_el_for_basis
        ) / CompTools(formula_pristine).stoich(shared_el_for_basis)

        pristine["multiplier"] = pristine_multiplier

        pristine_els = CompTools(pristine["formula"]).els
        pristine_amounts = [
            CompTools(pristine["formula"]).stoich(el) * pristine_multiplier
            for el in pristine_els
        ]
        pristine_basis_formula = ""
        for i, el in enumerate(pristine_els):
            pristine_basis_formula += el + str(int(pristine_amounts[i]))
        pristine["basis_formula"] = pristine_basis_formula

        self.pristine = pristine
        self.defect = defect
        self.gap_discretization = gap_discretization

    @property
    def dE(self):
        """
        Returns:
            the energy difference between the pristine structure and the defective structure
                E[X^q] - E[pristine]
            the basis is the defect-containing formula unit

            e.g.,
                if pristine = Al2O3
                    defect = Al4O5S1
                   E[X^q] = E_Al4O5S1
                   E[pristine] = 2*E_Al2O3 (to make Al4O6)
                   dE = E_Al4O5S1*10 atoms - 2*E_Al2O3*5 atoms

        """
        pristine = self.pristine
        defect = self.defect

        return (
            defect["E"] * CompTools(defect["formula"]).n_atoms
            - pristine["E"]
            * CompTools(pristine["formula"]).n_atoms
            * pristine["multiplier"]
        )

    @property
    def els_exchanged(self):
        """
        Returns:
            dictionary of elements exchanged between pristine and defect-containing compounds
            {el (str) :
                dn (float)}
                dn > 0 --> more of that el in defect than pristine
                dn < 0 --> less of that el in defect than pristine
        """
        pristine = self.pristine
        defect = self.defect
        els = sorted(
            list(
                set(
                    CompTools(pristine["formula"]).els
                    + CompTools(defect["formula"]).els
                )
            )
        )

        dn = {}
        for el in els:
            n_pristine = (
                CompTools(pristine["formula"]).stoich(el) * pristine["multiplier"]
            )
            n_defect = CompTools(defect["formula"]).stoich(el)
            dn[el] = n_defect - n_pristine

        return dn

    @property
    def defect_type(self):
        """
        Returns:
            (str) what type of defect is it?
                - vacancy
                - interstitial
                - substitional

            NOTE: what about antisite defects?
        """
        dn = self.els_exchanged
        removed_from_pristine = []
        added_to_pristine = []
        for el in dn:
            if dn[el] < 0:
                removed_from_pristine.append(el)
            elif dn[el] > 0:
                added_to_pristine.append(el)
        if (len(removed_from_pristine) > 0) and (len(added_to_pristine) > 0):
            defect_type = "substitional"
        elif len(removed_from_pristine) > 0:
            defect_type = "vacancy"
        elif len(added_to_pristine) > 0:
            defect_type = "interstitial"
        else:
            raise ValueError("defect_type not found")

        return defect_type

    @property
    def n_defects(self):
        """
        Returns:
            how many defects (int) (referenced to the defect-containing basis formula)
        """
        dn = self.els_exchanged
        el = [el for el in dn if dn[el] != 0][0]
        return abs(dn[el])

    @property
    def sum_mus(self):
        """
        Returns:
            sum(dn_i * mu_i) for all elements that are exchanged

            e.g., if pristine = Al2O3 and defect = Al4O5S1
                - so basis = Al4O6
                - dn_Al = 4 - 4 = 0
                - dn_O = 5 - 6 = -1
                - dn_S = 1 - 0 = 1
                - so sum_mus = -1 * mu_O + 1 * mu_S

        """
        chempots = self.chempots
        dn = self.els_exchanged
        sum_mus = 0.0
        for el in dn:
            if dn[el] == 0:
                continue
            sum_mus += dn[el] * chempots[el]

        return sum_mus

    @property
    def Efs(self):
        """

        Returns:
            {E_Fermi (float, eV relative to VBM) : defect formation energy (float, eV/defect)}

            Ef = Ef[X^q] - Ef[pristine] + sum(dn_i*mu_i) + q * E_Fermi + charge_correction

            note:
                - the charge correction gets divided by the number of defects
        """
        pristine = self.pristine
        Eg = pristine["Eg"]
        multiplier = pristine["multiplier"]
        gap_discretization = self.gap_discretization
        if Eg > 0:
            E_Fermis = np.arange(0, Eg + gap_discretization, gap_discretization)
        else:
            E_Fermis = np.array([0.0])
        dE = self.dE
        sum_mus = self.sum_mus
        q = self.defect["q"]
        charge_correction = self.charge_correction
        n_defects = self.n_defects
        energies = [
            ((dE - sum_mus) / multiplier + q * E_Fermi + charge_correction) / n_defects
            for E_Fermi in E_Fermis
        ]
        return dict(zip(E_Fermis, energies))


class ReactionEnergy(object):

    """
    *** This is a work in progress ***

    @TODO:
        - write tests/demo
        - incorporate filler
        - incorporate normalization
        - incorporate open systems

    """

    def __init__(
        self, input_energies, reactants, products, energy_key="Ef", norm="rxn"
    ):
        """

        Args:
            input_energies (dict)
                {formula (str): {< energy key> : formation energy (eV/at)}
                    formation energies should account for chemical potentials (e.g., due to partial pressures)

            reactants (list):
                list of reactant compositions (str)

            products (list):
                list of product compositions (str)

            energy_key (str):
                how to find the formation energies in formation_energies

            norm (str):
                how to normalize the reaction energy
                    'rxn' : the molar reaction energy for the reaction in ReactionEnergy.rxn_string
                    'atom' : the molar reaction energy per atom in the products side of the reaction in ReactionEnergy.rxn_string
        """

        self.input_energies = {
            CompTools(c).clean: {"E": input_energies[c][energy_key]}
            for c in input_energies
        }
        self.reactants = [Composition(c) for c in reactants]
        self.products = [Composition(c) for c in products]

    @property
    def rxn(self):
        """
        Returns:
            Pymatgen Reaction object
        """
        rxn = Reaction(self.reactants, self.products)
        return rxn

    @property
    def rxn_string(self):
        """
        Returns:
            string representation of the reaction
        """
        return self.rxn.__str__()

    @property
    def coefs(self):
        """
        Returns:
            {formula (str) : stoichiometry (float) in reaction}
        """
        rxn = self.rxn
        coefs = rxn._coeffs
        all_comp = rxn._all_comp
        all_comp = [CompTools(c.formula).clean for c in all_comp]
        unique_comp = list(set(all_comp))
        out = {c: 0 for c in unique_comp}
        for i, coef in enumerate(coefs):
            comp = all_comp[i]
            out[comp] += np.round(coef, 8)
        return out

    @property
    def species(self):
        """

        Returns:
            {formula (str) : {'coef' : stoichiometry (float) in reaction},
                              'E' : energy (float, eV/atom) of that formula}}
        """
        species = {}
        coefs = self.coefs
        energies = self.input_energies
        for c in coefs:
            if c != 0:
                species[c] = {
                    "coef": coefs[c],
                    "E": energies[c]["E"] if CompTools(c).n_els > 1 else 0,
                }

        return species

    @property
    def dE_rxn(self):
        """
        Returns:
            reaction energy (float)
                eV/atom if self.norm == 'atom'
                eV/rxn if self.norm == 'rxn'

        """
        species = self.species
        norm = self.norm
        dE_rxn = 0
        norm = 1
        for formula in species:
            coef = species[formula]["coef"]
            if (norm == "atom") and (coef > 0):
                norm += coef * CompTools(formula).n_atoms
            Ef = species[formula]["E"]
            dE_rxn += coef * Ef * CompTools(formula).n_atoms

        return dE_rxn / norm


def main():
    return


if __name__ == "__main__":
    out = main()



import multiprocessing as multip
import os
from pydmclab.hpc.launch import LaunchTools
from pydmclab.hpc.submit import SubmitTools
from pydmclab.hpc.analyze import AnalyzeVASP, AnalyzeBatch
from pydmclab.core.query import MPQuery
from pydmclab.core.struc import StrucTools
from pydmclab.core.mag import MagTools
from pydmclab.core.energies import ChemPots, FormationEnthalpy
from pydmclab.utils.handy import read_json, write_json


def get_vasp_configs(
    run_lobster=False,
    modify_loose_incar=False,
    modify_relax_incar=False,
    modify_static_incar=False,
    modify_loose_kpoints=False,
    modify_relax_kpoints=False,
    modify_static_kpoints=False,
    modify_loose_potcar=False,
    modify_relax_potcar=False,
    modify_static_potcar=False,
):
    """

    how to modify VASP calculations from the defaults (see pydmclab.hpc.vasp for defaults)

    Args:
        run_lobster (bool, optional): True to run LOBSTER

        modify_loose_incar (bool, optional):
            - dictionary of {incar flag (str) : setting for that flag}
            - modifies only the "loose" calculations
        modify_relax_incar (bool, optional):
            - dictionary of {incar flag (str) : setting for that flag}
            - modifies only the "relax" calculations
        modify_static_incar (bool, optional):
            - dictionary of {incar flag (str) : setting for that flag}
            - modifies only the "static" calculations


        modify_loose_kpoints (bool, optional):
            - dictionary of non-default K-point setting. see pydmclab.hpc.vasp for format
            - modifies only the "loose" calculations
        modify_relax_kpoints (bool, optional):
            - dictionary of non-default K-point setting. see pydmclab.hpc.vasp for format
            - modifies only the "relax" calculations
        modify_static_kpoints (bool, optional):
            - dictionary of non-default K-point setting. see pydmclab.hpc.vasp for format
            - modifies only the "static" calculations


        modify_loose_potcar (bool, optional):
            - dictionary of non-default POTCAR in format {element (str) : POTCAR to use (str)}
            - modifies only the "loose" calculations
        modify_relax_potcar (bool, optional):
            - dictionary of non-default POTCAR in format {element (str) : POTCAR to use (str)}
            - modifies only the "loose" calculations
        modify_static_potcar (bool, optional):
            - dictionary of non-default POTCAR in format {element (str) : POTCAR to use (str)}
            - modifies only the "static" calculations

    Returns:
        dictionary of VASP_CONFIGS
    """
    vasp_configs = {"lobster_static": run_lobster}

    if modify_loose_incar:
        vasp_configs["loose_incar"] = modify_loose_incar
    if modify_relax_incar:
        vasp_configs["relax_incar"] = modify_relax_incar
    if modify_static_incar:
        vasp_configs["static_incar"] = modify_static_incar

    if modify_loose_kpoints:
        vasp_configs["loose_kpoints"] = modify_loose_kpoints
    if modify_relax_kpoints:
        vasp_configs["relax_kpoints"] = modify_relax_kpoints
    if modify_static_kpoints:
        vasp_configs["static_kpoints"] = modify_static_kpoints

    if modify_loose_potcar:
        vasp_configs["loose_potcar"] = modify_loose_potcar
    if modify_relax_potcar:
        vasp_configs["relax_potcar"] = modify_relax_potcar
    if modify_static_potcar:
        vasp_configs["static_potcar"] = modify_static_potcar

    return vasp_configs


def get_slurm_configs(
    total_nodes=1,
    cores_per_node=8,
    walltime_in_hours=95,
    mem_per_core="all",
    partition="agsmall,msidmc",
    error_file="log.e",
    output_file="log.o",
    account="cbartel",
):
    """

    how to modify slurm configurations for each VASP job (see pydmclab.data.data._slurm_configs.yaml for defaults)

    Args:
        total_nodes (int, optional):
            how many nodes to run each VASP job on

        cores_per_node (int, optional):
            how many cores per node to use for each VASP job

        walltime_in_hours (int, optional):
            how long to run each VASP job

        mem_per_core (str, optional):
            if 'all', try to use all avaiable mem; otherwise use specified memory per core (cpu)

        partition (str, optional):
            what part of the cluster to run each VASP job on

        error_file (str, optional):
            where to send each VASP job error

        output_file (str, optional):
            where to send each VASP job output

        account (str, optional):
            what account to charge for your VASP jobs

    Returns:
        {slurm config name : slurm config value}
    """
    slurm_configs = {}

    slurm_configs["nodes"] = total_nodes
    slurm_configs["ntasks"] = int(total_nodes * cores_per_node)

    if account == "cbartel":
        slurm_configs["time"] = int(walltime_in_hours * 60)

    if total_nodes > 1:
        if "small" in partition:
            print("WARNING: cant use small partition on > 1 node; switching to large")
        partition = partition.replace("small", "large")

    slurm_configs["partition"] = partition

    slurm_configs["error_file"] = error_file
    slurm_configs["output_file"] = output_file
    slurm_configs["account"] = account

    if total_nodes > 4:
        print("WARNING: are you sure you need more than 4 nodes??")

    if (total_nodes > 1) and (cores_per_node < 32):
        print("WARNING: this seems like a small job. are you sure you need > 1 node??")

    if mem_per_core == "all":
        if partition in [
            "msismall",
            "small",
            "msilarge",
            "large",
            "amdsmall",
            "amdlarge",
            "RM-small",
            "RM",
        ]:
            mem_per_cpu = 1900
        elif partition in ["agsmall", "msidmc", "aglarge", "agsmall,msidmd", "RM-512"]:
            mem_per_cpu = 4000
        else:
            mem_per_cpu = 1900
    else:
        mem_per_cpu = mem_per_core

    slurm_configs["mem"] = str(int(mem_per_cpu * slurm_configs["ntasks"])) + "M"
    return slurm_configs


def get_sub_configs(
    machine="msi",
    submit_calculations_in_parallel=False,
    delete_all_calculations_and_start_over=False,
    rerun_lobster=False,
    mpi_command="mpirun",
    special_packing=False,
):
    """

    configs related to preparing submission scripts and submitting VASP calculations
        - see defaults in pydmclab.data.data._sub_configs.yaml

    Args:
        submit_calculations_in_parallel (bool or int): whether to prepare submission scripts in parallel or not
            - False: use 1 processor
            - True: use all available processors
            - int: use that many processors

        delete_all_calculations_and_start_over (bool):
            if True, start all calculations over (ie delete all outputs)

        rerun_lobster (bool) :
            if True, rerun lobster even if it has already been run

        mpi_command (str):
            the command to use for mpi (eg mpirun, srun, etc)

        special_packing (dict):
            if you want to change the loose --> relax --> static flow for some functional
                e.g., {'metagga' : ['loose', 'static']}

    Returns:
        {config_name : config_value}

    """
    sub_configs = {}

    if not submit_calculations_in_parallel:
        n_procs = 1
    else:
        if submit_calculations_in_parallel == True:
            n_procs = multip.cpu_count() - 1
        elif submit_calculations_in_parallel == False:
            n_procs = 1
        else:
            n_procs = submit_calculations_in_parallel

    sub_configs["n_procs"] = n_procs

    if delete_all_calculations_and_start_over:
        sub_configs["fresh_restart"] = True

    if rerun_lobster:
        sub_configs["force_postprocess"] = True

    sub_configs["mpi_command"] = mpi_command

    if special_packing:
        sub_configs["packing"] = {}
        for xc in special_packing:
            sub_configs["packing"]["xc"] = special_packing[xc]

    sub_configs["machine"] = machine

    return sub_configs


def get_launch_configs(
    standards=["dmc"],
    xcs=["metagga"],
    use_mp_thermo_data=False,
    n_afm_configs=0,
    skip_xcs_for_standards={"mp": ["gga", "metagga"]},
):
    """

    configs related to launching  chains of calculations

    Args:
        standards (list, optional):
            list of standards you'd like to calculate

        xcs (list, optional):
            list of xcs you'd like to calculate for each standard

        use_mp_thermo_data (bool, optional):
            True if you are going to use formation energies provided in Materials Project for phase stability analysis

        n_afm_configs (int, optional):
            number of antiferromagnetic configurations to run for each structure (0 if you don't want to run AFM)

        skip_xcs_for_standards (dict, optional):
            dictionary of xcs to skip for a given standard.
                Defaults to {"mp": ["gga", "metagga"]}.
                    - e.g., we don't want to run GGA or MetaGGA MP calculations because MP uses GGA+U (for now)

    Returns:
        dictionary of launch configurations
    """

    launch_configs = {}

    to_launch = {}

    for standard in standards:
        to_launch[standard] = xcs

    if use_mp_thermo_data:
        to_launch["mp"] = ["ggau"]

    for standard in skip_xcs_for_standards:
        if standard not in to_launch:
            continue
        for xc in skip_xcs_for_standards[standard]:
            if xc in to_launch[standard]:
                to_launch[standard].remove(xc)

    launch_configs["to_launch"] = to_launch

    launch_configs["compare_to_mp"] = use_mp_thermo_data

    launch_configs["n_afm_configs"] = n_afm_configs

    return launch_configs


def get_analysis_configs(
    analyze_calculations_in_parallel=False,
    analyze_structure=True,
    analyze_mag=False,
    analyze_charge=False,
    analyze_dos=False,
    analyze_bonding=False,
    exclude=[],
):
    """

    function for modifying analysis configs from the defaults (see pydmclab.data.data._batch_analysis_configs.yaml for defaults)

    Args:
        analyze_calculations_in_parallel (bool or int): whether to analyze calculation results in parallel or not
            - False: use 1 processor
            - True: use all available processors
            - int: use that many processors

        analyze_structure (bool, optional):
            True to include structure in your results

        analyze_mag (bool, optional):
            True to include magnetization in your results

        analyze_charge (bool, optional):
            True to include bader charge + lobster charges + madelung in your results

        analyze_dos (bool, optional):
            True to include pdos, tdos in your results

        analyze_bonding (bool, optional):
            True to include tcohp, pcohp, tcoop, pcoop, tcobi, pcobi in your results

        exclude (list, optional):
            list of strings to exclude from analysis. Defaults to [].
                - overwrites other options
    Returns:
        dictionary of ANALYSIS_CONFIGS
            {'include_*' : True or False}
    """

    analysis_configs = {}

    if not analyze_calculations_in_parallel:
        n_procs = 1
    else:
        if analyze_calculations_in_parallel == True:
            n_procs = multip.cpu_count() - 1
        elif analyze_calculations_in_parallel == False:
            n_procs = 1
        else:
            n_procs = analyze_calculations_in_parallel

    analysis_configs["n_procs"] = n_procs

    includes = []
    if analyze_structure:
        includes.append("structure")

    if analyze_mag:
        includes.append("mag")

    if analyze_charge:
        includes.extend(["charge", "madelung"])

    if analyze_dos:
        includes.extend(["tdos", "pdos"])

    if analyze_bonding:
        includes.extend(["tcohp", "pcohp", "tcoop", "pcoop", "tcobi", "pcobi"])

    for include in includes:
        analysis_configs["include_" + include] = True

    if exclude:
        for ex in exclude:
            analysis_configs["include_" + ex] = False

    return analysis_configs


def get_query(
    comp,
    api_key,
    properties=None,
    criteria=None,
    only_gs=True,
    include_structure=True,
    supercell_structure=False,
    max_Ehull=0.05,
    max_sites_per_structure=65,
    max_strucs_per_cmpd=4,
    data_dir=os.getcwd().replace("scripts", "data"),
    savename="query.json",
    remake=False,
):
    """
    Args:
        comp (list or str)
            can either be:
                - a chemical system (str) of elements joined by "-"
                - a chemical formula (str)
            can either be a list of:
                - chemical systems (str) of elements joined by "-"
                - chemical formulas (str)

        api_key (str):
            your API key for Materials Project

        properties (list or None)
            list of properties to query
                - if None, then use typical_properties
                - if 'all', then use supported_properties

        criteria (dict or None)
            dictionary of criteria to query
                - if None, then use {}

        only_gs (bool)
            if True, remove non-ground state polymorphs for each unique composition

        include_structure (bool)
            if True, include the structure (as a dictionary) for each entry

        supercell_structure (bool)
            only runs if include_structure = True
            if False, just retrieve the MP structure
            if not False, must be specified as [a,b,c] to make an a x b x c supercell of the MP structure

        max_Ehull (float)
            if not None, remove entries with Ehull_mp > max_Ehull

        max_sites_per_structure (int)
            if not None, remove entries with more than max_sites_per_structure sites

        max_strucs_per_cmpd (int)
            if not None, only retain the lowest energy structures for each composition until you reach max_strucs_per_cmpd

        data_dir (str)
            directory to save fjson

        savename (str)
            filename for fjson in data_dir

        remake (bool)
            write (True) or just read (False) fjson

    Returns:
        {ID (str) : {'structure' : Pymatgen Structure as dict,
                    < any other data you want to keep track of >}}
    """

    fjson = os.path.join(data_dir, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    # initialize MPQuery with your API key
    mpq = MPQuery(api_key=api_key)

    # get the data from MP
    data = mpq.get_data_for_comp(
        comp=comp,
        properties=properties,
        criteria=criteria,
        only_gs=only_gs,
        include_structure=include_structure,
        supercell_structure=supercell_structure,
        max_Ehull=max_Ehull,
        max_sites_per_structure=max_sites_per_structure,
        max_strucs_per_cmpd=max_strucs_per_cmpd,
    )

    write_json(data, fjson)
    return read_json(fjson)


def check_query(query):
    for mpid in query:
        print("\nmpid: %s" % mpid)
        print("\tcmpd: %s" % query[mpid]["cmpd"])
        print("\tstructure formula: %s" % StrucTools(query[mpid]["structure"]).formula)


def get_strucs(
    query,
    data_dir=os.getcwd().replace("scripts", "data"),
    savename="strucs.json",
    remake=False,
):
    """
    Args:
        query (dict)
            {mpid : {DATA}}

        data_dir (str)
            directory to save fjson

        savename (str)
            filename for fjson in DATA_DIR

        remake (bool)
            write (True) or just read (False) fjson

    Returns:
        {formula identifier (str) :
            {structure identifier for that formula (str) :
                Pymatgen Structure object as dict}}
    """

    fjson = os.path.join(data_dir, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    # get all unique chemical formulas in the query
    formulas_in_query = sorted(list(set([query[mpid]["cmpd"] for mpid in query])))

    data = {}
    for formula in formulas_in_query:
        # get all MP IDs in your query having that formula
        mpids = [mpid for mpid in query if query[mpid]["cmpd"] == formula]
        data[formula] = {mpid: query[mpid]["structure"] for mpid in mpids}

    write_json(data, fjson)
    return read_json(fjson)


def check_strucs(strucs):
    for formula in strucs:
        for ID in strucs[formula]:
            print("\nformula: %s" % formula)
            print("\tID: %s" % ID)
            struc = strucs[formula][ID]
            print("\tstructure formula: %s" % StrucTools(struc).formula)


def get_magmoms(
    strucs,
    max_afm_combos=50,
    treat_as_nm=[],
    data_dir=os.getcwd().replace("scripts", "data"),
    savename="magmoms.json",
    remake=False,
):
    """
    Args:
        strucs (dict)
            {formula : {ID : structure}}

        max_afm_combos (int)
            maximum number of AFM spin configurations to generate

        treat_as_nm (list)
            any normally mag els you'd like to treat as nonmagnetic for AFM enumeration

        data_dir (str)
            directory to save fjson

        savename (str)
            filename for fjson in data_dir

        remake (bool)
            write (True) or just read (False) fjson

    Returns:
        {formula identifier (str) :
            {structure identifier for that formula (str) :
                {AFM ordering identifier (str) :
                    [list of magmoms (floats) for each site in the structure]}}}"""

    fjson = os.path.join(data_dir, savename)
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    magmoms = {}
    for formula in strucs:
        magmoms[formula] = {}
        for ID in strucs[formula]:
            # for each unique structure, get AFM magmom orderings
            structure = strucs[formula][ID]
            magtools = MagTools(
                structure=structure,
                max_afm_combos=max_afm_combos,
                treat_as_nm=treat_as_nm,
            )
            curr_magmoms = magtools.get_afm_magmoms
            magmoms[formula][ID] = curr_magmoms

    write_json(magmoms, fjson)
    return read_json(fjson)


def check_magmoms(strucs, magmoms):
    for formula in strucs:
        for ID in strucs[formula]:
            structure_formula = StrucTools(strucs[formula][ID]).formula
            n_afm_configs = len(magmoms[formula][ID])
            print("%s: %i AFM configs\n" % (structure_formula, n_afm_configs))


def get_launch_dirs(
    strucs,
    magmoms,
    user_configs,
    make_launch_dirs=True,
    refresh_configs=True,
    data_dir=os.getcwd().replace("scripts", "data"),
    calcs_dir=os.getcwd().replace("scripts", "calcs"),
    savename="launch_dirs.json",
    remake=False,
):
    """
    Args:
        strucs (dict)
            {formula : {ID : structure}}

        magmoms (dict)
            {formula : {ID : {AFM configuration index : [list of magmoms on each site]}}

        user_configs (dict)
            optional launch configurations

        make_launch_dirs (bool)
            make launch directories (True) or just return launch dict (False)

        refresh_configs (bool)
            refresh configs (True) or just use existing configs (False)

        data_dir (str)
            directory to save fjson

        calcs_dir (str)
            directory above all your calculations

        savename (str)
            filename for fjson in data_dir

        remake (bool)
            write (True) or just read (False) fjson

    Returns:
        {launch dir (str, formula/ID/standard/mag) :
            {'xcs' : [list of final xcs to run (str)],
            'magmoms' : [list of magmoms (floats) for each site in the structure]}}

        also makes launch_dir and populates with POSCAR using strucs if make_dirs=True

    """

    fjson = os.path.join(data_dir, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    all_launch_dirs = {}
    for formula in strucs:
        for ID in strucs[formula]:
            # for each unique structure, generate our launch directories
            structure = strucs[formula][ID]
            if magmoms:
                curr_magmoms = magmoms[formula][ID]
            else:
                curr_magmoms = None
            top_level = formula
            unique_ID = ID

            launch = LaunchTools(
                calcs_dir=calcs_dir,
                user_configs=user_configs,
                magmoms=curr_magmoms,
                structure=structure,
                top_level=top_level,
                unique_ID=unique_ID,
                refresh_configs=refresh_configs,
            )

            launch_dirs = launch.launch_dirs(make_dirs=make_launch_dirs)

            for launch_dir in launch_dirs:
                all_launch_dirs[launch_dir] = launch_dirs[launch_dir]

    write_json(all_launch_dirs, fjson)
    return read_json(fjson)


def check_launch_dirs(launch_dirs):
    print("\nanalyzing launch directories")
    for d in launch_dirs:
        print("\nlaunching from %s" % d)
        print("   these final xcs: %s" % launch_dirs[d]["xcs"])


def submit_one_calc(submit_args):
    """
    Prepares VASP inputs, writes submission script, and launches job for one launch_dir

    Args:
        submit_args (dict) should contain:
            {'launch_dir' :
                launch_dir (str)
                    (formula/ID/standard/mag) to write and launch submission script in,
            'launch_dirs' :
                launch_dirs (dict)
                    {launch_dir (formula/ID/standard/mag) : {'xcs' : [list of final_xcs], 'magmoms' : [list of magmoms for each site in structure in launch_dir]}},
            'user_configs' :
                user_configs (dict)
                    optional sub, slurm, or VASP configurations,
            'refresh_configs' :
                refresh_configs (list)
                    list of which configs to refresh,
            'ready_to_launch':
                ready_to_launch (bool)
                    write and launch (True) or just write submission scripts (False)
                }

    Returns:
        None

    """
    launch_dir = submit_args["launch_dir"]
    launch_dirs = submit_args["launch_dirs"]
    user_configs = submit_args["user_configs"]
    refresh_configs = submit_args["refresh_configs"]
    ready_to_launch = submit_args["ready_to_launch"]

    # what are our terminal xcs for that launch_dir
    final_xcs = launch_dirs[launch_dir]["xcs"]

    # what magmoms apply to that launch_dir
    magmom = launch_dirs[launch_dir]["magmom"]

    try:
        sub = SubmitTools(
            launch_dir=launch_dir,
            final_xcs=final_xcs,
            magmom=magmom,
            user_configs=user_configs,
            refresh_configs=refresh_configs,
        )

        # prepare VASP directories and write submission script
        sub.write_sub

        # submit submission script to the queue
        if ready_to_launch:
            sub.launch_sub

        success = True
    except TypeError:
        print("\nERROR: %s\n   will submit without multiprocessing" % launch_dir)
        success = False

    return {"launch_dir": launch_dir, "success": success}


def submit_calcs(
    launch_dirs,
    user_configs={},
    refresh_configs=["vasp", "sub", "slurm"],
    ready_to_launch=True,
    n_procs=1,
):
    """
    Prepares VASP inputs, writes submission script, and launches job for all launch_dirs

    Args:
        launch_dirs (dict)
            {launch_dir (formula/ID/standard/mag) : {'xcs' : [list of final_xcs], 'magmoms' : [list of magmoms for each site in structure in launch_dir]}}

        user_configs (dict)
            optional sub, slurm, or VASP configurations

        refresh_configs (list)
            list of which configs to refresh

        ready_to_launch (bool)
            write and launch (True) or just write submission scripts (False

    Returns:
        None

    """

    submit_args = {
        "launch_dirs": launch_dirs,
        "user_configs": user_configs,
        "refresh_configs": refresh_configs,
        "ready_to_launch": ready_to_launch,
    }

    if n_procs == 1:
        print("\n\n submitting calculations in serial\n\n")
        for launch_dir in launch_dirs:
            curr_submit_args = submit_args.copy()
            curr_submit_args["launch_dir"] = launch_dir
            submit_one_calc(curr_submit_args)
        return
    elif n_procs == "all":
        n_procs = multip.cpu_count() - 1

    print("\n\n submitting calculations in parallel\n\n")
    print("not refreshing configs for parallel --> causes trouble")
    submit_args["refresh_configs"] = refresh_configs
    list_of_submit_args = []
    for launch_dir in launch_dirs:
        curr_submit_args = submit_args.copy()
        curr_submit_args["launch_dir"] = launch_dir
        list_of_submit_args.append(curr_submit_args)
    pool = multip.Pool(processes=n_procs)
    statuses = pool.map(submit_one_calc, list_of_submit_args)
    pool.close()

    submitted_w_multiprorcessing = [status for status in statuses if status["success"]]
    failed_w_multiprocessing = [status for status in statuses if not status["success"]]

    print(
        "%i/%i calculations submitted with multiprocessing"
        % (len(submitted_w_multiprorcessing), len(statuses))
    )
    for status in failed_w_multiprocessing:
        launch_dir = status["launch_dir"]
        curr_submit_args = submit_args.copy()
        curr_submit_args["launch_dir"] = launch_dir
        submit_one_calc(curr_submit_args)

    return


def get_results(
    launch_dirs,
    user_configs,
    refresh_configs=True,
    data_dir=os.getcwd().replace("scripts", "data"),
    savename="results.json",
    remake=False,
):
    """
    Args:
        launch_dirs (dict)
            {launch_dir (formula/ID/standard/mag) : {'xcs' : [list of final_xcs], 'magmoms' : [list of magmoms for each site in structure in launch_dir]}}

        user_configs (dict)
            optional analysis configurations

        refresh_configs (bool)
            refresh configs (True) or just use existing configs (False)

        data_dir (str)
            directory to save fjson

        savename (str)
            filename for fjson in data_dir

        remake (bool)
            write (True) or just read (False) fjson

    Returns:
        {formula--ID--standard--mag--xc-calc (str) :
            {scraped results from VASP calculation}}
    """

    fjson = os.path.join(data_dir, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    analyzer = AnalyzeBatch(
        launch_dirs, user_configs=user_configs, refresh_configs=refresh_configs
    )

    data = analyzer.results

    write_json(data, fjson)
    return read_json(fjson)


def check_results(results):
    keys_to_check = list(results.keys())

    converged = 0
    for key in keys_to_check:
        if "--" in key:
            delimiter = "--"
        else:
            delimiter = "."
        top_level, ID, standard, mag, xc_calc = key.split(delimiter)
        data = results[key]
        convergence = results[key]["results"]["convergence"]
        print("\n%s" % key)
        print("convergence = %s" % convergence)
        if convergence:
            converged += 1
            # print("\n%s" % key)
            print("E (static) = %.2f" % data["results"]["E_per_at"])

    print("\n\n SUMMARY: %i/%i converged" % (converged, len(keys_to_check)))


def get_gs(
    results,
    include_structure=False,
    non_default_functional=None,
    compute_Ef=True,
    data_dir=os.getcwd().replace("scripts", "data"),
    savename="gs.json",
    remake=False,
):
    """
    Args:
        results (dict)
            {formula--ID--standard--mag--xc-calc (str) : {scraped results from VASP calculation}}

        include_structure (bool)
            include the structure or not

        non_default_functional (str)
            if you're not using r2SCAN or PBE

        compute_Ef (bool)
            if True, compute formation enthalpy

        data_dir (str)
            directory to save fjson

        savename (str)
            filename for fjson in data_dir

        remake (bool)
            write (True) or just read (False) fjson

    Returns:
    {standard (str, the calculation standard) :
        {xc (str, the exchange-correlation method) :
            {formula (str) :
                {'E' : energy of the ground-structure,
                'key' : formula.ID.standard.mag.xc_calc for the ground-state structure,
                'structure' : structure of the ground-state structure,
                'n_started' : how many polymorphs you tried to calculate,
                'n_converged' : how many polymorphs are converged,
                'complete' : True if n_converged = n_started (i.e., all structures for this formula at this xc are done),
                'Ef' : formation enthalpy at 0 K}
    """
    fjson = os.path.join(data_dir, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    standards = sorted(
        list(set([results[key]["meta"]["setup"]["standard"] for key in results]))
    )

    gs = {
        standard: {
            xc: {}
            for xc in sorted(
                list(
                    set(
                        [
                            results[key]["meta"]["setup"]["xc"]
                            for key in results
                            if results[key]["meta"]["setup"]["standard"] == standard
                        ]
                    )
                )
            )
        }
        for standard in standards
    }

    for standard in gs:
        for xc in gs[standard]:
            keys = [
                k
                for k in results
                if results[k]["meta"]["setup"]["standard"] == standard
                if results[k]["meta"]["setup"]["xc"] == xc
            ]

            unique_formulas = sorted(
                list(set([results[key]["results"]["formula"] for key in keys]))
            )
            for formula in unique_formulas:
                gs[standard][xc][formula] = {}
                formula_keys = [
                    k for k in keys if results[k]["results"]["formula"] == formula
                ]
                converged_keys = [
                    k for k in formula_keys if results[k]["results"]["convergence"]
                ]
                if not converged_keys:
                    gs_energy, gs_structure, gs_key = None, None, None
                else:
                    energies = [
                        results[k]["results"]["E_per_at"] for k in converged_keys
                    ]
                    gs_energy = min(energies)
                    gs_key = converged_keys[energies.index(gs_energy)]
                    gs_structure = results[gs_key]["structure"]
                complete = True if len(formula_keys) == len(converged_keys) else False
                gs[standard][xc][formula] = {
                    "E": gs_energy,
                    "key": gs_key,
                    "n_started": len(formula_keys),
                    "n_converged": len(converged_keys),
                    "complete": complete,
                }
                if include_structure:
                    gs[standard][xc][formula]["structure"] = gs_structure

    if compute_Ef:
        for standard in gs:
            for xc in gs[standard]:
                if not non_default_functional:
                    functional = "r2scan" if xc == "metagga" else "pbe"
                else:
                    functional = non_default_functional
                mus = ChemPots(functional=functional, standard=standard).chempots
                for formula in gs[standard][xc]:
                    E = gs[standard][xc][formula]["E"]
                    if E:
                        Ef = FormationEnthalpy(
                            formula=formula, E_DFT=E, chempots=mus
                        ).Ef
                    else:
                        Ef = None
                    gs[standard][xc][formula]["Ef"] = Ef
    write_json(gs, fjson)
    return read_json(fjson)


def check_gs(gs):
    """
    checks that this dictionary is generated properly

    Args:
        gs (_type_): _description_

    """

    print("\nchecking ground-states")
    standards = gs.keys()
    print("standards = ", standards)
    for standard in standards:
        print("\nworking on %s standard" % standard)
        xcs = list(gs[standard].keys())
        for xc in xcs:
            print("  xc = %s" % xc)
            formulas = list(gs[standard][xc].keys())
            n_formulas = len(formulas)
            n_formulas_complete = len(
                [k for k in formulas if gs[standard][xc][k]["complete"]]
            )
            print(
                "%i/%i formulas with all calculations completed"
                % (n_formulas_complete, n_formulas)
            )
            for formula in gs[standard][xc]:
                if "Ef" in gs[standard][xc][formula]:
                    if gs[standard][xc][formula]["Ef"]:
                        print(
                            "%s : %.2f eV/at"
                            % (formula, gs[standard][xc][formula]["Ef"])
                        )


def get_thermo_results(
    results,
    gs,
    data_dir=os.getcwd().replace("scripts", "data"),
    savename="thermo_results.json",
    remake=False,
):
    """

    Args:
        results (dict):
            full results dictionary

        gs (dict):
            dictionary of ground-state data

        data_dir (str)
            directory to save fjson

        savename (str)
            fjson name in data_dir

        remake (bool)
            Read (False) or write (True) json

    Returns:
        {standard (str) :
            {xc (str) :
                {formula (str) :
                    {ID (str) :
                        {'E' : energy of the structure (DFT total energy in eV/atom),
                        'Ef' : formation enthalpy at 0 K (eV/atom),
                        'is_gs' : True if this is the lowest energy polymorph for this formula,
                        'dE_gs' : how high above the ground-state this structure is in energy (eV/atom)
                        'all_polymorphs_converged' : True if every structure that was computed for this formula is converged}}
    """
    fjson = os.path.join(data_dir, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    thermo_results = {
        standard: {
            xc: {formula: {} for formula in gs[standard][xc]} for xc in gs[standard]
        }
        for standard in gs
    }

    for key in results:
        tmp_thermo = {}

        standard = results[key]["meta"]["setup"]["standard"]
        xc = results[key]["meta"]["setup"]["xc"]
        formula = results[key]["results"]["formula"]
        ID = "__".join(
            [
                results[key]["meta"]["setup"]["formula_tag"],
                results[key]["meta"]["setup"]["ID"],
                results[key]["meta"]["setup"]["mag"],
            ]
        )
        E = results[key]["results"]["E_per_at"]
        formula = results[key]["results"]["formula"]
        structure = results[key]["structure"]
        if structure:
            calcd_formula = StrucTools(structure).formula
        else:
            calcd_formula = None
        tmp_thermo["E"] = E
        tmp_thermo["key"] = key
        tmp_thermo["formula"] = formula
        tmp_thermo["calculated_formula"] = calcd_formula

        if E:
            gs_key = gs[standard][xc][formula]["key"]
            if "Ef" in gs[standard][xc][formula]:
                gs_Ef = gs[standard][xc][formula]["Ef"]
            else:
                gs_Ef = None
            gs_E = gs[standard][xc][formula]["E"]
            delta_E_gs = E - gs_E

            if key == gs_key:
                tmp_thermo["is_gs"] = True
            else:
                tmp_thermo["is_gs"] = False

            tmp_thermo["dE_gs"] = delta_E_gs
            if gs_Ef:
                tmp_thermo["Ef"] = gs_Ef + delta_E_gs
            else:
                tmp_thermo["Ef"] = None
            tmp_thermo["all_polymorphs_converged"] = gs[standard][xc][formula][
                "complete"
            ]

        else:
            tmp_thermo["dE_gs"] = None
            tmp_thermo["Ef"] = None
            tmp_thermo["is_gs"] = False
            tmp_thermo["all_polymorphs_converged"] = False

        thermo_results[standard][xc][formula][ID] = tmp_thermo

    write_json(thermo_results, fjson)
    return read_json(fjson)


def check_thermo_results(thermo):
    print("\nchecking thermo results")

    for standard in thermo:
        print("\n\nworking on %s standard" % standard)
        for xc in thermo[standard]:
            print("\nxc = %s" % xc)
            for formula in thermo[standard][xc]:
                print("formula = %s" % formula)
                print(
                    "%i polymorphs converged"
                    % len(
                        [
                            k
                            for k in thermo[standard][xc][formula]
                            if thermo[standard][xc][formula][k]["E"]
                        ]
                    )
                )
                gs_ID = [
                    k
                    for k in thermo[standard][xc][formula]
                    if thermo[standard][xc][formula][k]["is_gs"]
                ]
                if gs_ID:
                    gs_ID = gs_ID[0]
                    print("%s is the ground-state structure" % gs_ID)

    print("\n\n  SUMMARY  ")
    for standard in thermo:
        for xc in thermo[standard]:
            print("~~%s~~" % "_".join([standard, xc]))
            converged_formulas = []
            for formula in thermo[standard][xc]:
                for ID in thermo[standard][xc][formula]:
                    if thermo[standard][xc][formula][ID]["all_polymorphs_converged"]:
                        converged_formulas.append(formula)

            converged_formulas = list(set(converged_formulas))
            print(
                "%i/%i formulas have all polymorphs converged"
                % (len(converged_formulas), len(thermo[standard][xc].keys()))
            )


def get_dos_results(
    results,
    thermo_results,
    only_gs=True,
    only_xc="metagga",
    only_formulas=[],
    only_standard="dmc",
    dos_to_store=["tdos", "tcohp"],
    regenerate_dos=False,
    regenerate_cohp=False,
    data_dir=os.getcwd().replace("scripts", "data"),
    savename="dos_results.json",
    remake=False,
):
    """
    Args:
        results (dict)
            from get_results

        thermo_results (dict)
            from get_thermo_results

        only_gs (bool)
            if True, only get DOS/COHP for the ground-state polymorphs

        only_xc (str)
            if not None, only get DOS/COHP for this XC

        only_formulas (list)
            if not None, only get DOS/COHP for these formulas

        only_standard (str)
            if not None, only get DOS/COHP for this standard

        dos_to_store (list)
            which DOS/COHP to store ['tcohp', 'pcohp', 'tdos', 'pdos', etc]

        regenerate_dos (bool)
            if True, make pdos/tdos jsons again even if it exists

        regenerate_cohp (bool)
            if True, make pcohp/tcohp jsons again even if it exists

        data_dir (str)
            path to data directory

        savename (str)
            name of json file to save results to

        remake (bool)
            if True, remake the json file
    """
    fjson = os.path.join(data_dir, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    for key in results:
        calc_dir = results[key]["meta"]["calc_dir"]
        standard, xc = (
            results[key]["meta"]["setup"]["standard"],
            results[key]["meta"]["setup"]["xc"],
        )
        ID = "__".join(
            [
                results[key]["meta"]["setup"]["formula_tag"],
                results[key]["meta"]["setup"]["ID"],
                results[key]["meta"]["setup"]["mag"],
            ]
        )
        formula = results[key]["results"]["formula"]
        thermo_result = thermo_results[standard][xc][formula][ID]
        if only_gs:
            if not thermo_result["is_gs"]:
                continue
        if only_formulas:
            if thermo_result["formula"] not in only_formulas:
                continue
        if only_xc:
            if xc != only_xc:
                continue
        if only_standard:
            if standard != only_standard:
                continue
        av = AnalyzeVASP(calc_dir)
        if "tdos" in dos_to_store:
            pdos = av.pdos(remake=regenerate_dos)
            tdos = av.tdos(pdos=pdos, remake=regenerate_dos)
            thermo_results[standard][xc][formula][ID]["tdos"] = tdos
        if "pdos" in dos_to_store:
            thermo_results[standard][xc][formula][ID]["pdos"] = pdos
        if "tcohp" in dos_to_store:
            pcohp = av.pcohp(remake=regenerate_cohp)
            tcohp = av.tcohp(pcohp=pcohp, remake=regenerate_cohp)
            thermo_results[standard][xc][formula][ID]["tcohp"] = tcohp
        if "pcohp" in dos_to_store:
            thermo_results[standard][xc][formula][ID]["pcohp"] = pcohp
        if "tcoop" in dos_to_store:
            pcohp = av.pcohp(are_coops=True, remake=regenerate_cohp)
            tcohp = av.tcohp(pcohp=pcohp, remake=regenerate_cohp)
            thermo_results[standard][xc][formula][ID]["tcoop"] = tcohp
        if "pcoop" in dos_to_store:
            thermo_results[standard][xc][formula][ID]["pcoop"] = pcohp
        if "tcobi" in dos_to_store:
            pcohp = av.pcohp(are_cobis=True, remake=regenerate_cohp)
            tcohp = av.tcohp(pcohp=pcohp, remake=regenerate_cohp)
            thermo_results[standard][xc][formula][ID]["tcobi"] = tcohp
        if "pcobi" in dos_to_store:
            thermo_results[standard][xc][formula][ID]["pcobi"] = pcohp

    write_json(thermo_results, fjson)
    return read_json(fjson)


def crawl_and_purge(
    head_dir,
    files_to_purge=[
        "WAVECAR",
        "CHGCAR",
        "CHG",
        "PROCAR",
        "LOCPOT",
        "AECCAR0",
        "AECCAR1",
        "AECCAR2",
    ],
    safety="on",
    check_convergence=True,
    verbose=False,
):
    """
    Args:
        head_dir (str)
            directory to start crawling beneath

        files_to_purge (list)
            list of file names to purge

        safety (str)
            'on' or 'off' to turn on/off safety
                - if safety is on, won't actually delete files
    """
    purged_files = []
    mem_created = 0
    for subdir, dirs, files in os.walk(head_dir):
        ready = False
        if check_convergence:
            if "POTCAR" in files:
                av = AnalyzeVASP(subdir)
                if av.is_converged:
                    ready = True
                else:
                    ready = False
            else:
                ready = False
        else:
            ready = True
        if ready:
            for f in files:
                if f in files_to_purge:
                    path_to_f = os.path.join(subdir, f)
                    if verbose:
                        print(path_to_f)
                    mem_created += os.stat(path_to_f).st_size
                    purged_files.append(path_to_f)
                    if safety == "off":
                        os.remove(path_to_f)
    if safety == "off":
        print(
            "You purged %i files, freeing up %.2f GB of memory"
            % (len(purged_files), mem_created / 1e9)
        )
    if safety == "on":
        print(
            "You had the safety on\n If it were off, you would have purged %i files, freeing up %.2f GB of memory"
            % (len(purged_files), mem_created / 1e9)
        )


def make_sub_for_launcher():
    """
    Creates sub_launcher.sh file to launch launcher on compute node
    """
    flauncher_sub = os.path.join(os.getcwd(), "sub_launcher.sh")
    launch_job_name = "-".join([os.getcwd().split("/")[-2], "launcher"])
    with open(flauncher_sub, "w") as f:
        f.write("#!/bin/bash -l\n")
        f.write("#SBATCH --nodes=1\n")
        f.write("#SBATCH --ntasks=8\n")
        f.write("#SBATCH --time=4:00:00\n")
        f.write("#SBATCH --mem=8G\n")
        f.write("#SBATCH --error=_log_launcher.e\n")
        f.write("#SBATCH --output=_log_launcher.o\n")
        f.write("#SBATCH --account=cbartel\n")
        f.write("#SBATCH --job-name=%s\n" % launch_job_name)
        f.write("#SBATCH --partition=msismall\n")
        f.write("\npython launcher.py\n")


def main():
    mus = ChemPots(functional="r2scan", standard="dmc").chempots
    for el in mus:
        if not mus[el]:
            print(el)


if __name__ == "__main__":
    main()
